
/*
 * @(#)AccessBridge.java	1.50 04/05/17
 * 
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package com.sun.java.accessibility;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.lang.*;
import java.lang.reflect.*;

import java.beans.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

import javax.accessibility.*;
import com.sun.java.accessibility.util.*;

/**
 * 1.50 04/05/17
 * @author      Peter Korn
 * @author      Lynn Monsanto
 */
public class AccessBridge implements Runnable {

    public static final String AccessBridgeVersion = 
	"AccessBridge 1.2 GA";

    /**
     * Load JavaAccessBridge.DLL (our native half)
     */
    static {
	java.security.AccessController.doPrivileged(
	    new java.security.PrivilegedAction() {
		public Object run() {
		    System.loadLibrary("JavaAccessBridge");
		    return null;
		}
	    });
    }

    /**
     * main
     */
    static public void main(String args[]) {
        new AccessBridge();
    }

    private static AccessBridge theAccessBridge;
    private ObjectReferences references;
    private EventHandler eventHandler;
    private boolean runningOnJDK1_4 = false;
    private boolean runningOnJDK1_5 = false;
    private boolean useJAWT_DLL = false;

    // Maps AccessibleRoles strings to AccessibleRoles.
    // We shouldn't have to do this since it requires us
    // to synchronize the allAccessibleRoles array when
    // the AccessibleRoles class interface changes.  However,
    // there is no method to get all AccessibleRoles
    private Hashtable accessibleRoleMap = new Hashtable();

    /**
     * AccessBridge constructor
     */
    public AccessBridge() {
	theAccessBridge = this;
	references = new ObjectReferences();

	// initialize AccessibleRoles map
	for (int i = 0; i < allAccessibleRoles.length; i++) {
	    AccessibleRole nextRole = allAccessibleRoles[i];
	    String nextRoleString = nextRole.toDisplayString(Locale.US);
	    debugString("role string = "+nextRoleString+
			"; role = "+nextRole);
	    accessibleRoleMap.put(nextRoleString, nextRole);
	}

	// determine which version of the JDK is running
	String version = getJavaVersionProperty();
	debugString("JDK version = "+version);
	useJAWT_DLL = (version.compareTo("1.4.1") >= 0);
	runningOnJDK1_4 = (version.compareTo("1.4") >= 0);
	runningOnJDK1_5 = (version.compareTo("1.5") >= 0);

	// load JAWTAccessBridge.DLL on JDK 1.4.1 or later
	if (useJAWT_DLL) {
	    // Note that we have to explicitly load JAWT.DLL
	    java.security.AccessController.doPrivileged(
	        new java.security.PrivilegedAction() {
		    public Object run() {
			System.loadLibrary("JAWT");
			System.loadLibrary("JAWTAccessBridge");
			return null;
		    }
		});
	}
	
	// initialize the methods that map HWNDs and Java top-level 
	// windows
        if (initHWNDcalls() == true) {  

            // is this a JVM we can use?
	    // install JDK 1.2 and later Swing ToolKit listener
            EventQueueMonitor.isGUIInitialized();   

	    // start the Java event handler
            eventHandler = new EventHandler(this);

	    // register for menu selection events
	    if (runningOnJDK1_4) {
		MenuSelectionManager.defaultManager().addChangeListener(eventHandler);
	    }

	    // register as a NativeWindowHandler
	    addNativeWindowHandler(new DefaultNativeWindowHandler());

	    // start in a new thread
            new Thread(this).start();
            debugString("AccessBridge started");
        } else {
            debugString("Cannot load the AccessBridge - couldn't init HWND calls");
        }
    }
    
    /**
     * start the AccessBridge DLL running in its own thread
     */
    public native void run();

    /**
     * debugging output (goes to OutputDebugStr())
     */
    public native void sendDebugString(String debugStr);

    /**
     * debugging output (goes to OutputDebugStr())
     */
    public void debugString(String debugStr) {
	// System.err.println(debugStr);
	sendDebugString(debugStr);
    }

    /* ===== utility methods ===== */

    /**
     * decrement the reference to the object (called by native code)
     */
    public void decrementReference(Object o) {
	references.decrement(o);
    }

    /**
     * get the java.version property from the JVM
     */
    public String getJavaVersionProperty() {
        String s = System.getProperty("java.version");
	if (s != null) {
	    references.increment(s);
	    return s;
        }
        return null;
    }

    /**
     * get the java.version property from the JVM
     */
    public String getAccessBridgeVersion() {
        String s = new String(AccessBridgeVersion);
	references.increment(s);
	return s;
    }

    /* ===== HWND/Java window mapping methods ===== */

    // Java toolkit methods for mapping HWNDs to Java components
    private Method javaGetComponentFromNativeWindowHandleMethod;
    private Method javaGetNativeWindowHandleFromComponentMethod;

    // native jawt methods for mapping HWNDs to Java components
    public native int isJAWTInstalled();

    public native int jawtGetNativeWindowHandleFromComponent(Component comp);

    public native Component jawtGetComponentFromNativeWindowHandle(int handle);

    Toolkit toolkit;

    /**
     * map an HWND to an AWT Component
     */
    public boolean initHWNDcalls() {
	Class integerParemter[] = new Class[1];
        integerParemter[0] = Integer.TYPE;
	Class componentParemter[] = new Class[1];
        try {
            componentParemter[0] = Class.forName("java.awt.Component");
        } catch (ClassNotFoundException e) {
            debugString("Exception: " + e.toString());
        }
        Object[] args = new Object[1];
        Component c;
        boolean returnVal = false;
	
	toolkit = Toolkit.getDefaultToolkit();

	if (useJAWT_DLL) {
	    returnVal = true;
	} else {
	    // verify javaGetComponentFromNativeWindowHandle() method
	    // is present if JAWT.DLL is not installed
	    try {
		javaGetComponentFromNativeWindowHandleMethod = toolkit.getClass().getMethod(
											    "getComponentFromNativeWindowHandle", integerParemter);
		if (javaGetComponentFromNativeWindowHandleMethod != null) {
		    try {
			args[0] = new Integer(1);
			c = (Component) javaGetComponentFromNativeWindowHandleMethod.invoke(toolkit, args);
			returnVal = true;
		    } catch (InvocationTargetException e) {
			debugString("Exception: " + e.toString());
		    } catch (IllegalAccessException e) {
			debugString("Exception: " + e.toString());
		    }
		}
		
	    } catch (NoSuchMethodException e) {
		debugString("Exception: " + e.toString());
	    } catch (SecurityException e) {
		debugString("Exception: " + e.toString());
	    }
	    
	    // verify getComponentFromNativeWindowHandle() method 
	    // is present if JAWT.DLL is not installed
	    try {
		javaGetNativeWindowHandleFromComponentMethod = toolkit.getClass().getMethod(
											    "getNativeWindowHandleFromComponent", componentParemter);
		if (javaGetNativeWindowHandleFromComponentMethod != null) {
		    try {
			args[0] = new Button("OK");    // need some Component...
			Integer i = (Integer) javaGetNativeWindowHandleFromComponentMethod.invoke(toolkit, args);
			returnVal = true;
		    } catch (InvocationTargetException e) {
			debugString("Exception: " + e.toString());
		    } catch (IllegalAccessException e) {
			debugString("Exception: " + e.toString());
		    } catch (Exception e) {
			debugString("Exception: " + e.toString());
		    }
		}
		
	    } catch (NoSuchMethodException e) {
		debugString("Exception: " + e.toString());
	    } catch (SecurityException e) {
		debugString("Exception: " + e.toString());
	    }
	}
        return returnVal;
    }

    // native window handler interface
    public interface NativeWindowHandler {
	public Accessible getAccessibleFromNativeWindowHandle(int nativeHandle);
    }

    // hash table of native window handle to AccessibleContext mappings
    static private Hashtable windowHandleToContextMap = new Hashtable();

    // hash table of AccessibleContext to native window handle mappings
    static private Hashtable contextToWindowHandleMap = new Hashtable();

    /*
     * adds a virtual window handler to our hash tables
     */
    static public void registerVirtualFrame( Accessible a, 
					     Integer nativeWindowHandle ) {
	if (a != null) {
	    windowHandleToContextMap.put(nativeWindowHandle, 
					 a.getAccessibleContext());

	    contextToWindowHandleMap.put(a.getAccessibleContext(), 
					 nativeWindowHandle);
	}
    }

    /*
     * removes a virtual window handler to our hash tables
     */
    static public void revokeVirtualFrame( Accessible a, 
					   Integer nativeWindowHandle ) {
	windowHandleToContextMap.remove(nativeWindowHandle);
	contextToWindowHandleMap.remove(a.getAccessibleContext());
    }

    // vector of native window handlers
    private static Vector nativeWindowHandlers = new Vector();
    
    /*
     * adds a native window handler to our list
     */
    public static void addNativeWindowHandler(NativeWindowHandler handler) {
	theAccessBridge.debugString("addNativeWindowHandler="+handler.getClass());
	if (handler == null) {
	    throw new IllegalArgumentException();
	}
	nativeWindowHandlers.addElement(handler);
    }

    /*
     * removes a native window handler to our list
     */
    public static boolean removeNativeWindowHandler(NativeWindowHandler handler) {
	theAccessBridge.debugString("removeNativeWindowHandler="+handler.getClass());
	if (handler == null) {
	    throw new IllegalArgumentException();
	}
	return nativeWindowHandlers.removeElement(handler);
    }

    /**
     * verifies that a native window handle is a Java window
     */
    public boolean isJavaWindow(int nativeHandle) {
        AccessibleContext ac = getContextFromNativeWindowHandle(nativeHandle);
        if (ac != null) {
	    saveContextToWindowHandleMapping(ac, nativeHandle);
            return true;
        }
	return false;
    }

    /*
     * saves the mapping between an AccessibleContext and a window handle
     */
    private void saveContextToWindowHandleMapping(AccessibleContext ac,
						  int nativeHandle) {
	debugString("saveContextToWindowHandleMapping...");
	if (ac == null) {
	    return;
	}
	if (! contextToWindowHandleMap.containsKey(ac)) {
	    debugString("saveContextToWindowHandleMapping: ac = "+ac+"; handle = "+nativeHandle);
	    contextToWindowHandleMap.put(ac, new Integer(nativeHandle));
	}
    }

    /**
     * maps a native window handle to an Accessible Context
     */
    public AccessibleContext getContextFromNativeWindowHandle(int nativeHandle) {
	// First, look for the Accessible in our hash table of
	// virtual window handles.
	Object o = windowHandleToContextMap.get(new Integer(nativeHandle));
	if (o instanceof AccessibleContext) {
	    saveContextToWindowHandleMapping((AccessibleContext)o, nativeHandle);
	    return (AccessibleContext)o;
	}

	// Next, look for the native window handle in our vector
	// of native window handles.
	int numHandlers = nativeWindowHandlers.size();
	for (int i = 0; i < numHandlers; i++) {
	    NativeWindowHandler nextHandler = 
		(NativeWindowHandler)nativeWindowHandlers.elementAt(i);
	    Accessible a = 
		nextHandler.getAccessibleFromNativeWindowHandle(nativeHandle);
	    if (a != null) {
		saveContextToWindowHandleMapping(a.getAccessibleContext(), nativeHandle);
		return a.getAccessibleContext();
	    }
	}
	// Not found.
	return null;
    }

    /**
     * maps an AccessibleContext to a native window handle
     *     returns 0 on error
     */
    public int getNativeWindowHandleFromContext(AccessibleContext ac) {
	debugString("getNativeWindowHandleFromContext: ac = "+ac);
	Enumeration enum = contextToWindowHandleMap.keys();
	while (enum.hasMoreElements()) {
	    debugString(" next elem = "+enum.nextElement());
	}
	Object o = contextToWindowHandleMap.get(ac);
	if (o instanceof Integer) {
	    debugString(" returning (success) "+ ((Integer)o).intValue());
	    return ((Integer)o).intValue();
	}
	debugString(" returning (failure) 0");
	return 0;
    }

    protected class DefaultNativeWindowHandler implements NativeWindowHandler {
	/*
	 * returns the Accessible associated with a native window
	 */
	public Accessible getAccessibleFromNativeWindowHandle(int nativeHandle) {
	    Component c = getComponentFromNativeWindowHandle(nativeHandle);
	    if (c instanceof Accessible) {
		saveContextToWindowHandleMapping(((Accessible)c).getAccessibleContext(), 
						 nativeHandle);
		return (Accessible)c;
	    } else {
		return null;
	    }
	}
	
	/**
	 * map an HWND to an AWT Component
	 */
	protected Component getComponentFromNativeWindowHandle(int nativeHandle) {
	    if (useJAWT_DLL) {
		debugString("*** calling jawtGetComponentFromNativeWindowHandle");
		return jawtGetComponentFromNativeWindowHandle(nativeHandle);
	    } else {
		debugString("*** calling javaGetComponentFromNativeWindowHandle");
		Object[] args = new Object[1];
		if (javaGetComponentFromNativeWindowHandleMethod != null) {
		    try {
			args[0] = new Integer(nativeHandle);
			Object o = 
			    javaGetComponentFromNativeWindowHandleMethod.invoke(toolkit, args);
			if (o instanceof Accessible) {
			    saveContextToWindowHandleMapping(((Accessible)o).getAccessibleContext(), 
							     nativeHandle);
			}
			return (Component)o;
		    } catch (InvocationTargetException e) {
			debugString("Exception: " + e.toString());
		    } catch (IllegalAccessException e) {
			debugString("Exception: " + e.toString());
		    }
		}
	    }
	    return null;
	}
    }

    /**
     * map an AWT Component to an HWND
     */
    public int getNativeWindowHandleFromComponent(Component target) {
	
	if (useJAWT_DLL) {
	    debugString("*** calling jawtGetNativeWindowHandleFromComponent");
	    return jawtGetNativeWindowHandleFromComponent(target);
	    
	} else {
	    Object[] args = new Object[1];
	    debugString("*** calling javaGetNativeWindowHandleFromComponent");
	    if (javaGetNativeWindowHandleFromComponentMethod != null) {
		try {
		    args[0] = target;
		    Integer i = (Integer) javaGetNativeWindowHandleFromComponentMethod.invoke(toolkit, args);
		    // cache the mapping
		    contextToWindowHandleMap.put(target.getAccessibleContext(), i);

		    return i.intValue();
		} catch (InvocationTargetException e) {
		    debugString("Exception: " + e.toString());
		} catch (IllegalAccessException e) {
		    debugString("Exception: " + e.toString());
		}
	    }
	}
        return -1;
    }
    
    /* ===== AccessibleContext methods =====*/

    /*
     * returns the inner-most AccessibleContext in parent at Point(x, y)
     */
    public AccessibleContext getAccessibleContextAt(int x, int y, 
						    AccessibleContext parent) {
	
	if (parent == null) {
	    return null;
	}
	
	if (windowHandleToContextMap != null && 
	    windowHandleToContextMap.containsValue(getRootAccessibleContext(parent))) {
	    // Path for applications that register their top-level
	    // windows with the AccessBridge (e.g., StarOffice 6.1)
	    return getAccessibleContextAt_1(x, y, parent);
	} else {
	    // Path for applications that do not register
	    // their top-level windows with the AccessBridge
	    // (e.g., Swing/AWT applications)
	    return getAccessibleContextAt_2(x, y, parent);
	}
    }
    
    /*
     * returns the root accessible context
     */
    private AccessibleContext getRootAccessibleContext(AccessibleContext ac) {
	if (ac == null) {
	    return null;
	}
	Accessible parent = ac.getAccessibleParent();
	if (parent == null) {
	    return ac;
	}
	Accessible tmp = parent.getAccessibleContext().getAccessibleParent();
	while (tmp != null) {
	    parent = tmp;
	    tmp = parent.getAccessibleContext().getAccessibleParent();
	}
	return parent.getAccessibleContext();
    }
    
    /*
     * StarOffice version that does not use the EventQueueMonitor
     */
    public AccessibleContext getAccessibleContextAt_1(int x, int y, 
						      AccessibleContext parent) {
	debugString(" : getAccessibleContextAt_1 called");
	debugString("   -> x = " + x + " y = " + y + " parent = " + parent);
	
	if (parent != null) {
	    AccessibleComponent acmp = parent.getAccessibleComponent();
	    if (acmp != null) {
		Point loc = acmp.getLocation();
		Accessible a = acmp.getAccessibleAt(new Point(x - loc.x, y - loc.y));
		if (a != null) {
		    AccessibleContext foundAC = a.getAccessibleContext();
		    if (foundAC != null) {
			if (foundAC != parent) {
			    // recurse down into the child
			    return getAccessibleContextAt_1(x - loc.x, y - loc.y, 
				foundAC);
			} else {
			    debugString("   returning ac = " + foundAC);
			    return foundAC;
			}
		    }
		}
	    }
	}
	return parent;
    }
    
    /*
     * AWT/Swing version
     */
    public AccessibleContext getAccessibleContextAt_2(int x, int y, 
						      AccessibleContext parent) {
	debugString("getAccessibleContextAt_2 called");
	debugString("   -> x = " + x + " y = " + y + " parent = " + parent);
	
	Accessible a = EventQueueMonitor.getAccessibleAt(new Point(x, y));
	if (a != null) {
	    AccessibleContext childAC = a.getAccessibleContext();
	    if (childAC != null) {
		debugString("   returning childAC = " + childAC);
		return childAC;
	    }
	}
	debugString("returning AccessibleContext = null");
	return null;
    }
    
    /**
     * returns the Accessible that has focus
     */
    public AccessibleContext getAccessibleContextWithFocus() {
	Component c = AWTEventMonitor.getComponentWithFocus();
	if (c != null) {
	    Accessible a = Translator.getAccessible(c);
	    if (a != null) {
		AccessibleContext ac = a.getAccessibleContext();
		if (ac != null) {
		    return ac;
		}
	    }
	}
	return null;
    }
    
    /**
     * returns the AccessibleName from an AccessibleContext
     */
    public String getAccessibleNameFromContext(AccessibleContext ac) {
	debugString("***** ac = "+ac.getClass());
	if (ac != null) {
	    String s = ac.getAccessibleName();
	    if (s != null) {
		references.increment(s);
                debugString("Returning AccessibleName from Context: " + s);
		return s;
	    } else {
		return null;
            }
	} else {
	    debugString("getAccessibleNameFromContext; ac = null!");
	    return null;
        }
    }
    
    /**
     * returns the AccessibleDescription from an AccessibleContext
     */
    public String getAccessibleDescriptionFromContext(AccessibleContext ac) {
	if (ac != null) {
	    String s = ac.getAccessibleDescription();
	    if (s != null) {
		references.increment(s);
                debugString("Returning AccessibleDescription from Context: " + s);
		return s;
            }
        } else {
	    debugString("getAccessibleDescriptionFromContext; ac = null");
	}
	return null;
    }
    
    /**
     * returns the AccessibleRole from an AccessibleContext
     */
    public String getAccessibleRoleStringFromContext(AccessibleContext ac) {
	if (ac != null) {
	    AccessibleRole role = ac.getAccessibleRole();
            if (role != null) {
                String s = role.toDisplayString();
		if (s != null) {
		    references.increment(s);
                    debugString("Returning AccessibleRole from Context: " + s);
		    return s;
                }
            }
	} else {
	    debugString("getAccessibleRoleStringFromContext; ac = null");
	}
        return null;
    }

    /**
     * return the AccessibleRole from an AccessibleContext in the en_US locale
     */
    public String getAccessibleRoleStringFromContext_en_US(AccessibleContext ac) {
	if (ac != null) {
	    AccessibleRole role = ac.getAccessibleRole();
            if (role != null) {
                String s = role.toDisplayString(Locale.US);
		if (s != null) {
		    references.increment(s);
                    debugString("Returning AccessibleRole in en_US from Context: " + s);
		    return s;
                }
            }
	} else {
	    debugString("getAccessibleRoleStringFromContext; ac = null");
	}
        return null;
    }

    /**
     * return the AccessibleStates from an AccessibleContext
     */
    public String getAccessibleStatesStringFromContext(AccessibleContext ac) {
	if (ac != null) {
	    AccessibleStateSet stateSet = ac.getAccessibleStateSet();
            if (stateSet != null) {
                String s = stateSet.toString();
		if (s != null) {
		    // Indicate whether this component manages its own
		    // children
		    AccessibleRole role = ac.getAccessibleRole();
		    if (role == AccessibleRole.TABLE ||
			role == AccessibleRole.TREE) {
			s += ", manages_children";
		    }
		    references.increment(s);
                    debugString("Returning AccessibleStateSet from Context: " + s);
		    return s;
                }
            }
	} else {
	    debugString("getAccessibleStatesStringFromContext; ac = null");
	}
	return null;
    }

    /**
     * returns the AccessibleStates from an AccessibleContext in the en_US locale
     */
    public String getAccessibleStatesStringFromContext_en_US(AccessibleContext ac) {
	if (ac != null) {
	    AccessibleStateSet stateSet = ac.getAccessibleStateSet();
            if (stateSet != null) {
                String s = "";
                AccessibleState[] states = stateSet.toArray();
                if (states != null && states.length > 0) {
                    s = states[0].toDisplayString(Locale.US);
                    for (int i = 1; i < states.length; i++) {
                        s = s + "," + states[i].toDisplayString(Locale.US);
                    }
                }
                references.increment(s);
                debugString("Returning AccessibleStateSet en_US from Context: " + s);
                return s;
            }
	}
    	debugString("getAccessibleStatesStringFromContext; ac = null");
	return null;
    }

    /**
     * returns the AccessibleParent from an AccessibleContext
     */
    public AccessibleContext getAccessibleParentFromContext(AccessibleContext ac) {
	if (ac != null) {
	    Accessible a = ac.getAccessibleParent();
            if (a != null) {
                AccessibleContext apc = a.getAccessibleContext();
		if (apc != null) {
                    debugString("Returning AccessibleParent from Context");
		    return apc;
                }
            }
	} else {
	    debugString("getAccessibleParentFromContext; ac = null");
	}
	return null;
    }

    /**
     * returns the AccessibleIndexInParent from an AccessibleContext
     */
    public int getAccessibleIndexInParentFromContext(AccessibleContext ac) {
	if (ac != null) {
            debugString("Returning Accessible IndexInParent from Context: " + ac.getAccessibleIndexInParent());
	    return ac.getAccessibleIndexInParent();
	} else {
	    debugString("AccessibleIndexInParent; ac = null");
	}
	return -1;
    }

    /**
     * returns the AccessibleChild count from an AccessibleContext
     */
    public int getAccessibleChildrenCountFromContext(AccessibleContext ac) {
	if (ac != null) {
            debugString("Returning Accessible ChildrenCount from Context: " + ac.getAccessibleChildrenCount());
	    return ac.getAccessibleChildrenCount();
	} else {
	    debugString("getAccessibleChildrenCountFromContext; ac = null");
	}
	return -1;
    }

    /**
     * returns the AccessibleChild Context from an AccessibleContext
     */
    public AccessibleContext getAccessibleChildFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            Accessible a = ac.getAccessibleChild(index);
            if (a != null) {
                AccessibleContext acc = a.getAccessibleContext();
                if (acc != null) { 
                    debugString("Returning AccessibleChild from Context");
                    return acc;
                }
            } else {
		//        		debugString("getAccessibleChildFromContext; a = null");
            }
	} else {
	    debugString("getAccessibleChildFromContext; ac = null");
	}
	return null;
    }

    /**
     * returns the AccessibleComponent bounds on screen from an AccessibleContext
     */
    public Rectangle getAccessibleBoundsOnScreenFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleComponent acmp = ac.getAccessibleComponent();
            if (acmp != null) {
                Rectangle r = acmp.getBounds();
                if (r != null) {
                    try {
                        Point p = acmp.getLocationOnScreen();
                        if (p != null) {
                            r.x = p.x;
                            r.y = p.y;
                            return r;
                        }
                    } catch (Exception e) {
                        debugString("*** Caught EXCEPTION in getAccessibleBoundsOnScreenFromContext: " + e.toString());
                    }
                }
            }
	} else {
	    debugString("getAccessibleBoundsOnScreenFromContext; ac = null");
	}
	return null;
    }

    /**
     * returns the AccessibleComponent x-coord from an AccessibleContext
     */
    public int getAccessibleXcoordFromContext(AccessibleContext ac) {
	if (ac != null) {
            Rectangle r = getAccessibleBoundsOnScreenFromContext(ac);
            if (r != null) {
                debugString(" - Returning Accessible x coord from Context: " + r.x);
                return r.x;
            }
	} else {
	    debugString("getAccessibleXcoordFromContext ac = null");
	}
	return -1;
    }

    /**
     * returns the AccessibleComponent y-coord from an AccessibleContext
     */
    public int getAccessibleYcoordFromContext(AccessibleContext ac) {
        debugString("getAccessibleYcoordFromContext() called");
	if (ac != null) {
            Rectangle r = getAccessibleBoundsOnScreenFromContext(ac);
            if (r != null) {
                return r.y;
            }
	} else {
	    debugString("getAccessibleYcoordFromContext; ac = null");
	}
	return -1;
    }

    /**
     * returns the AccessibleComponent height from an AccessibleContext
     */
    public int getAccessibleHeightFromContext(AccessibleContext ac) {
	if (ac != null) {
            Rectangle r = getAccessibleBoundsOnScreenFromContext(ac);
            if (r != null) {
                return r.height;
            }
	} else {
	    debugString("getAccessibleHeightFromContext; ac = null");
	}
	return -1;
    }

    /**
     * returns the AccessibleComponent width from an AccessibleContext
     */
    public int getAccessibleWidthFromContext(AccessibleContext ac) {
	if (ac != null) {
            Rectangle r = getAccessibleBoundsOnScreenFromContext(ac);
            if (r != null) {
                return r.width;
            }
	} else {
	    debugString("getAccessibleWidthFromContext; ac = null");
	}
	return -1;
    }


    /**
     * returns the AccessibleComponent from an AccessibleContext
     */
    public AccessibleComponent getAccessibleComponentFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleComponent acmp = ac.getAccessibleComponent();
            if (acmp != null) {
                debugString("Returning AccessibleComponent Context");
                return acmp;
            }
	} else {
	    debugString("getAccessibleComponentFromContext; ac = null");
	}
	return null;
    }

    /**
     * returns the AccessibleAction from an AccessibleContext
     */
    public AccessibleAction getAccessibleActionFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleAction aa = ac.getAccessibleAction();
            if (aa != null) {
                debugString("Returning AccessibleAction Context");
                return aa;
            }
	} else {
	    debugString("getAccessibleActionFromContext; ac = null");
	}
	return null;
    }

    /**
     * returns the AccessibleSelection from an AccessibleContext
     */
    public AccessibleSelection getAccessibleSelectionFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                debugString("returning AccessibleSelection Context="+as.getClass());
                return as;
            }
	} else {
	    debugString("getAccessibleSelectionFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the AccessibleText from an AccessibleContext
     */
    public AccessibleText getAccessibleTextFromContext(AccessibleContext ac) {
        if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                debugString("Returning AccessibleText Context");
                return at;
            }
	} else {
	    debugString("getAccessibleTextFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the AccessibleComponent from an AccessibleContext
     */
    public AccessibleValue getAccessibleValueFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleValue av = ac.getAccessibleValue();
            if (av != null) {
                debugString("Returning AccessibleValue Context");
                return av;
            } else {
            }
	} else {
	    debugString("getAccessibleValueFromContext; ac = null");
	}
	return null;
    }

    /* ===== AccessibleText methods ===== */

    /**
     * returns the character count from an AccessibleContext
     */
    public int getAccessibleCharCountFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                return at.getCharCount();
            }
	} else {
	    debugString("getAccessibleCharCountFromContext; ac = null");
	}
	return -1;
    }

    /**
     * returns the caret position from an AccessibleContext
     */
    public int getAccessibleCaretPositionFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                return at.getCaretPosition();
            }
	} else {
	    debugString("getAccessibleCaretPositionFromContext; ac = null");
	}
	return -1;
    }

    /**
     * Return the index at a specific point from an AccessibleContext
     * Point(x, y) is in screen coordinates.
     */
    public int getAccessibleIndexAtPointFromContext(AccessibleContext ac, 
						    int x, int y) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
	    AccessibleComponent acomp = ac.getAccessibleComponent();
            if (at != null && acomp != null) {
		// Convert x and y from screen coordinates to 
		// local coordinates.
		try {
		    Point p = acomp.getLocationOnScreen();
		    if (p != null) {
			x = x - p.x;
			if (x < 0) {
			    x = 0;
			}
			y = y - p.y;
			if (y < 0) {
			    y = 0;
			}
			return at.getIndexAtPoint(new Point(x, y));
		    }
		} catch (Exception e) {
		}
            }
	}
	return -1;
    }

    /**
     * return the letter at a specific point from an AccessibleContext
     */
    public String getAccessibleLetterAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                String s = at.getAtIndex(AccessibleText.CHARACTER, index);
                if (s != null) {
                    references.increment(s);
                    return s;
                }
            }
	} else {
	    debugString("getAccessibleLetterAtIndexFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the word at a specific point from an AccessibleContext
     */
    public String getAccessibleWordAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                String s = at.getAtIndex(AccessibleText.WORD, index);
                if (s != null) {
                    references.increment(s);
                    return s;
                }
            }
	} else {
	    debugString("getAccessibleWordAtIndexFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the sentence at a specific point from an AccessibleContext
     */
    public String getAccessibleSentenceAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                String s = at.getAtIndex(AccessibleText.SENTENCE, index);
                if (s != null) {
                    references.increment(s);
                    return s;
                }
            }
	} else {
	    debugString("getAccessibleSentenceAtIndexFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the text selection start from an AccessibleContext
     */
    public int getAccessibleTextSelectionStartFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                return at.getSelectionStart();
            }
	} else {
	    debugString("getAccessibleTextSelectionStartFromContext; ac = null");
	}
	return -1;
    }

    /**
     * return the text selection end from an AccessibleContext
     */
    public int getAccessibleTextSelectionEndFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                return at.getSelectionEnd();
            }
	} else {
	    debugString("getAccessibleTextSelectionEndFromContext; ac = null");
	}
	return -1;
    }

    /**
     * return the selcted text from an AccessibleContext
     */
    public String getAccessibleTextSelectedTextFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                String s = at.getSelectedText();
                if (s != null) {
                    references.increment(s);
                    return s;
                }
            }
	} else {
	    debugString("getAccessibleTextSelectedTextFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the attribute string at a given index from an AccessibleContext
     */
    public String getAccessibleAttributesAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                AttributeSet as = at.getCharacterAttribute(index);
		//                String s = expandAttributes(as);
                String s = expandStyleConstants(as);
                if (s != null) {
                    references.increment(s);
                    return s;
                }
            }
	} else {
	    debugString("getAccessibleAttributesAtIndexFromContext; ac = null");
	}
	return null;
    }

    /**
     * Get line info: left index of line
     *
     * algorythm:  cast back, doubling each time,
     *             'till find line boundaries
     *              
     * return -1 if we can't get the info (e.g. index or at passed in
     * is bogus; etc.) 
     */
    public int getAccessibleTextLineLeftBoundsFromContext(AccessibleContext ac, int index) {
        if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                int lineStart;
                int offset;
                Rectangle charRect;
                Rectangle indexRect = at.getCharacterBounds(index);
                int textLen = at.getCharCount();

                if (indexRect == null) {
                    return -1;
                }

                // find the start of the line
                //
                offset = 1;
                lineStart = index - offset < 0 ? 0 : index - offset;
                charRect = at.getCharacterBounds(lineStart);

		// slouch behind beginning of line
                while (charRect != null
                       && charRect.y >= indexRect.y 
                       && lineStart > 0) {
                    offset = offset << 1;
                    lineStart = index - offset < 0 ? 0 : index - offset;
                    charRect = at.getCharacterBounds(lineStart);
                }
                if (lineStart == 0) {    // special case: we're on the first line!
                    // we found it!
                } else {
                    offset = offset >> 1;   // know boundary within last expansion

		    // ground forward to beginning of line
                    while (offset > 0) {
                        charRect = at.getCharacterBounds(lineStart + offset);
                        if (charRect.y < indexRect.y) { // still before line
                            lineStart += offset;
                        } else {
                            // leave lineStart alone, it's close!
                        }
                        offset = offset >> 1;
                    }
		    // subtract one 'cause we're already too far...
                    lineStart += 1;
                }
                return lineStart;
            }
	} else {
	    debugString("getAccessibleTextLineLeftBoundsFromContext; ac = null");
	}
	return -1;
    }


    /**
     * Get line info: right index of line
     *
     * algorithm:  cast back, doubling each time,
     *             'till find line boundaries
     *              
     * return -1 if we can't get the info (e.g. index or at passed in
     * is bogus; etc.) 
     */
    public int getAccessibleTextLineRightBoundsFromContext(AccessibleContext ac, int index) {
        if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                int lineEnd;
                int offset;
                Rectangle charRect;
                Rectangle indexRect = at.getCharacterBounds(index);
                int textLen = at.getCharCount();

                if (indexRect == null) {
                    return -1;
                }

                // find the end of the line
                //
                offset = 1;
                lineEnd = index + offset > textLen - 1 
		    ? textLen - 1 : index  + offset;
                charRect = at.getCharacterBounds(lineEnd);

		// push past end of line
                while (charRect != null 
                       && charRect.y <= indexRect.y 
                       && lineEnd < textLen - 1) {
                    offset = offset << 1;
                    lineEnd = index + offset > textLen - 1 
			? textLen - 1 : index + offset;
                    charRect = at.getCharacterBounds(lineEnd);
                }
                if (lineEnd == textLen - 1) {    // special case: on the last line!
                    // we found it!
                } else {
                    offset = offset >> 1;   // know boundary within last expansion

		    // pull back to end of line
                    while (offset > 0) {
                        charRect = at.getCharacterBounds(lineEnd - offset);
                        if (charRect.y > indexRect.y) { // still beyond line
                            lineEnd -= offset;
                        } else {
                            // leave lineEnd alone, it's close!
                        }
                        offset = offset >> 1;
                    }
		    // subtract one 'cause we're already too far...
                    lineEnd -= 1;
                }

                return lineEnd;
            }
	} else {
	    debugString("getAccessibleTextLineRightBoundsFromContext; ac = null");
	}
	return -1;
    }


    /**
     * Get a range of text; null if indicies are bogus
     */
    public String getAccessibleTextRangeFromContext(AccessibleContext ac, int start, int end) {
        if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                if (start >= end) {
                    return null;
                }
                if (end >= at.getCharCount()) {
                    return null;
                }
                StringBuffer buf = new StringBuffer(end-start+1);
                for (int i = start; i <= end; i++) {
                    buf.append(at.getAtIndex(AccessibleText.CHARACTER, i));
                }
                String s = buf.toString();
                if (s != null) {
                    references.increment(s);
                    return s;
                }
            }
	} else {
	    debugString("getAccessibleTextRangeFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the AttributeSet object at a given index from an AccessibleContext
     */
    public AttributeSet getAccessibleAttributeSetAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                AttributeSet as = at.getCharacterAttribute(index);
                if (as != null) {
		    references.increment(as);
                    return as;
                }
            }
	} else {
	    debugString("getAccessibleAttributeSetAtIndexFromContext; ac = null");
	}
	return null;
    }


    /**
     * return the bounding rectangle at index from an AccessibleContext
     */
    public Rectangle getAccessibleTextRectAtIndexFromContext(AccessibleContext ac, int index) {
        // want to do this in global coords, so need to combine w/ac global coords
	if (ac != null) {
            AccessibleText at = ac.getAccessibleText();
            if (at != null) {
                Rectangle r = at.getCharacterBounds(index);
                if (r != null) {
		    String s = at.getAtIndex(AccessibleText.CHARACTER, index);
		    if (s != null && s.equals("\n")) {
		        r.width = 0;
		    }
                    Rectangle acRect = getAccessibleBoundsOnScreenFromContext(ac);
                    if (acRect != null) {
                        r.translate(acRect.x, acRect.y);
                        return r;
                    }
                }
            }
	} else {
	    debugString("getAccessibleTextRectAtIndexFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the AccessibleText character x-coord at index from an AccessibleContext
     */
    public int getAccessibleXcoordTextRectAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            Rectangle r = getAccessibleTextRectAtIndexFromContext(ac, index);
            if (r != null) {
                return r.x;
            }
	} else {
	    debugString("getAccessibleXcoordTextRectAtIndexFromContext; ac = null");
	}
	return -1;
    }

    /**
     * return the AccessibleText character y-coord at index from an AccessibleContext
     */
    public int getAccessibleYcoordTextRectAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            Rectangle r = getAccessibleTextRectAtIndexFromContext(ac, index);
            if (r != null) {
                return r.y;
            }
	} else {
	    debugString("getAccessibleYcoordTextRectAtIndexFromContext; ac = null");
	}
	return -1;
    }

    /**
     * return the AccessibleText character height at index from an AccessibleContext
     */
    public int getAccessibleHeightTextRectAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            Rectangle r = getAccessibleTextRectAtIndexFromContext(ac, index);
            if (r != null) {
                return r.height;
            }
	} else {
	    debugString("getAccessibleHeightTextRectAtIndexFromContext; ac = null");
	}
	return -1;
    }

    /**
     * return the AccessibleText character width at index from an AccessibleContext
     */
    public int getAccessibleWidthTextRectAtIndexFromContext(AccessibleContext ac, int index) {
	if (ac != null) {
            Rectangle r = getAccessibleTextRectAtIndexFromContext(ac, index);
            if (r != null) {
                return r.width;
            }
	} else {
	    debugString("getAccessibleWidthTextRectAtIndexFromContext; ac = null");
	}
	return -1;
    }

    /* ===== AttributeSet methods for AccessibleText ===== */

    /**
     * return the bold setting from an AttributeSet
     */
    public boolean getBoldFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.isBold(as);
	} else {
	    debugString("getBoldFromAttributeSet; as = null");
	}
	return false;
    }

    /**
     * return the italic setting from an AttributeSet
     */
    public boolean getItalicFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.isItalic(as);
	} else {
	    debugString("getItalicFromAttributeSet; as = null");
	}
	return false;
    }

    /**
     * return the underline setting from an AttributeSet
     */
    public boolean getUnderlineFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.isUnderline(as);
	} else {
	    debugString("getUnderlineFromAttributeSet; as = null");
	}
	return false;
    }

    /**
     * return the strikethrough setting from an AttributeSet
     */
    public boolean getStrikethroughFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.isStrikeThrough(as);
	} else {
	    debugString("getStrikethroughFromAttributeSet; as = null");
	}
	return false;
    }

    /**
     * return the superscript setting from an AttributeSet
     */
    public boolean getSuperscriptFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.isSuperscript(as);
	} else {
	    debugString("getSuperscriptFromAttributeSet; as = null");
	}
	return false;
    }

    /**
     * return the subscript setting from an AttributeSet
     */
    public boolean getSubscriptFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.isSubscript(as);
	} else {
	    debugString("getSubscriptFromAttributeSet; as = null");
	}
	return false;
    }

    /**
     * return the background color from an AttributeSet
     */
    public String getBackgroundColorFromAttributeSet(AttributeSet as) {
	if (as != null) {
            String s = StyleConstants.getBackground(as).toString();
            if (s != null) {
                references.increment(s);
                return s;
            }
	} else {
	    debugString("getBackgroundColorFromAttributeSet; as = null");
	}
	return null;
    }

    /**
     * return the foreground color from an AttributeSet
     */
    public String getForegroundColorFromAttributeSet(AttributeSet as) {
	if (as != null) {
            String s = StyleConstants.getForeground(as).toString();
            if (s != null) {
                references.increment(s);
                return s;
            }
	} else {
	    debugString("getForegroundColorFromAttributeSet; as = null");
	}
	return null;
    }

    /**
     * return the font family from an AttributeSet
     */
    public String getFontFamilyFromAttributeSet(AttributeSet as) {
	if (as != null) {
            String s = StyleConstants.getFontFamily(as).toString();
            if (s != null) {
                references.increment(s);
                return s;
            }
	} else {
	    debugString("getFontFamilyFromAttributeSet; as = null");
	}
	return null;
    }

    /**
     * return the font size from an AttributeSet
     */
    public int getFontSizeFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getFontSize(as);
	} else {
	    debugString("getFontSizeFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the alignment from an AttributeSet
     */
    public int getAlignmentFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getAlignment(as);
	} else {
	    debugString("getAlignmentFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the BiDi level from an AttributeSet
     */
    public int getBidiLevelFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getBidiLevel(as);
	} else {
	    debugString("getBidiLevelFromAttributeSet; as = null");
	}
	return -1;
    }


    /**
     * return the first line indent from an AttributeSet
     */
    public float getFirstLineIndentFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getFirstLineIndent(as);
	} else {
	    debugString("getFirstLineIndentFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the left indent from an AttributeSet
     */
    public float getLeftIndentFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getLeftIndent(as);
	} else {
	    debugString("getLeftIndentFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the right indent from an AttributeSet
     */
    public float getRightIndentFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getRightIndent(as);
	} else {
	    debugString("getRightIndentFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the line spacing from an AttributeSet
     */
    public float getLineSpacingFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getLineSpacing(as);
	} else {
	    debugString("getLineSpacingFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the space above from an AttributeSet
     */
    public float getSpaceAboveFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getSpaceAbove(as);
	} else {
	    debugString("getSpaceAboveFromAttributeSet; as = null");
	}
	return -1;
    }

    /**
     * return the space below from an AttributeSet
     */
    public float getSpaceBelowFromAttributeSet(AttributeSet as) {
	if (as != null) {
            return StyleConstants.getSpaceBelow(as);
	} else {
	    debugString("getSpaceBelowFromAttributeSet; as = null");
	}
	return -1;
    }


    /**
     * Enumerate all StyleConstants in the AttributeSet
     *
     * We need to check explicitly, 'cause of the HTML package conversion
     * mechanism (they may not be stored as StyleConstants, just translated
     * to them when asked).
     *
     * (Use convenience methods where they are defined...)
     *
     * Not checking the following (which the IBM SNS guidelines says 
     * should be defined):
     *    - ComponentElementName
     *    - IconElementName
     *    - NameAttribute
     *    - ResolveAttribute
     */
    private String expandStyleConstants(AttributeSet as) {
	Color c;
	Object o;
	String attrString = new String("");

	// ---------- check for various Character Constants

	attrString += "BidiLevel = " + StyleConstants.getBidiLevel(as);

	Component comp = StyleConstants.getComponent(as);
	if (comp != null) {
	    if (comp instanceof Accessible) {
		AccessibleContext ac = ((Accessible) comp).getAccessibleContext();
		if (ac != null) {
		    attrString += "; Accessible Component = " + ac.getAccessibleName();
		} else {
		    attrString += "; Innaccessible Component = " + comp;
		}
	    } else {
		attrString += "; Innaccessible Component = " + comp;
	    }
	}

	Icon i = StyleConstants.getIcon(as);
	if (i != null) {
	    if (i instanceof ImageIcon) {
		attrString += "; ImageIcon = " + ((ImageIcon) i).getDescription();
	    } else {
		attrString += "; Icon = " + i;
	    }
	}

	attrString += "; FontFamily = " + StyleConstants.getFontFamily(as);

	attrString += "; FontSize = " + StyleConstants.getFontSize(as);

	if (StyleConstants.isBold(as)) {
	    attrString += "; bold";
	}

	if (StyleConstants.isItalic(as)) {
	    attrString += "; italic";
	}

	if (StyleConstants.isUnderline(as)) {
	    attrString += "; underline";
	}

	if (StyleConstants.isStrikeThrough(as)) {
	    attrString += "; strikethrough";
	}

	if (StyleConstants.isSuperscript(as)) {
	    attrString += "; superscript";
	}

	if (StyleConstants.isSubscript(as)) {
	    attrString += "; subscript";
	}

	c = StyleConstants.getForeground(as);
	if (c != null) {
	    attrString += "; Foreground = " + c;
	}

	c = StyleConstants.getBackground(as);
	if (c != null) {
	    attrString += "; Background = " + c;
	}

	attrString += "; FirstLineIndent = " + StyleConstants.getFirstLineIndent(as);

	attrString += "; RightIndent = " + StyleConstants.getRightIndent(as);

	attrString += "; LeftIndent = " + StyleConstants.getLeftIndent(as);

	attrString += "; LineSpacing = " + StyleConstants.getLineSpacing(as);

	attrString += "; SpaceAbove = " + StyleConstants.getSpaceAbove(as);

	attrString += "; SpaceBelow = " + StyleConstants.getSpaceBelow(as);

	attrString += "; Alignment = " + StyleConstants.getAlignment(as);

	TabSet ts = StyleConstants.getTabSet(as);
	if (ts != null) {
	    attrString += "; TabSet = " + ts; 
        }

	return attrString;
    }


    /* ===== AccessibleValue methods ===== */

    /**
     * return the AccessibleValue current value from an AccessibleContext
     * returned using a String 'cause the value is a java Number
     *
     */
    String getCurrentAccessibleValueFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleValue av = ac.getAccessibleValue();
            if (av != null) {
                Number value = av.getCurrentAccessibleValue();
                if (value != null) {
                    String s = value.toString();
                    if (s != null) {
                        references.increment(s);
                        return s;
                    }
                }
            }
	} else {
	    debugString("getCurrentAccessibleValueFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the AccessibleValue maximum value from an AccessibleContext
     * returned using a String 'cause the value is a java Number
     *
     */
    String getMaximumAccessibleValueFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleValue av = ac.getAccessibleValue();
            if (av != null) {
                Number value = av.getMaximumAccessibleValue();
                if (value != null) {
                    String s = value.toString();
                    if (s != null) {
                        references.increment(s);
                        return s;
                    }
                }
            }
	} else {
	    debugString("getMaximumAccessibleValueFromContext; ac = null");
	}
	return null;
    }

    /**
     * return the AccessibleValue minimum value from an AccessibleContext
     * returned using a String 'cause the value is a java Number
     *
     */
    String getMinimumAccessibleValueFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleValue av = ac.getAccessibleValue();
            if (av != null) {
                Number value = av.getMinimumAccessibleValue();
                if (value != null) {
                    String s = value.toString();
                    if (s != null) {
                        references.increment(s);
                        return s;
                    }
                }
            }
	} else {
	    debugString("getMinimumAccessibleValueFromContext; ac = null");
	}
	return null;
    }


    /* ===== AccessibleSelection methods ===== */

    /**
     * add to the AccessibleSelection of an AccessibleContext child i
     *
     */
    void addAccessibleSelectionFromContext(AccessibleContext ac, int i) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                as.addAccessibleSelection(i);
            }
	} else {
	    debugString("addAccessibleSelectionFromContext; ac = null");
	}
    }

    /**
     * clear all of the AccessibleSelection of an AccessibleContex
     *
     */
    void clearAccessibleSelectionFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                as.clearAccessibleSelection();
            }
	} else {
	    debugString("clearAccessibleSelectionFromContext; ac = null");
	}
    }

    /**
     * get the AccessibleContext of the i-th AccessibleSelection of an AccessibleContext
     *
     */
    AccessibleContext getAccessibleSelectionFromContext(AccessibleContext ac, int i) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                Accessible a = as.getAccessibleSelection(i);
                if (a != null) {
                    return a.getAccessibleContext();
                }
            }
	} else {
	    debugString("getAccessibleSelectionFromContext; ac = null");
	}
        return null;
    }

    /**
     * get number of things selected in the AccessibleSelection of an AccessibleContext
     *
     */
    int getAccessibleSelectionCountFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                return as.getAccessibleSelectionCount();
            }
	} else {
	    debugString("getAccessibleSelectionFromContext; ac = null");
	}
        return -1;
    }

    /**
     * return true if the i-th child of the AccessibleSelection of an AccessibleContext is selected
     *
     */
    boolean isAccessibleChildSelectedFromContext(AccessibleContext ac, int i) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                return as.isAccessibleChildSelected(i);
            }
	} else {
	    debugString("isAccessibleChildSelectedFromContext; ac = null");
	}
        return false;
    }

    /**
     * remove the i-th child from the AccessibleSelection of an AccessibleContext
     *
     */
    void removeAccessibleSelectionFromContext(AccessibleContext ac, int i) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                as.removeAccessibleSelection(i);
            }
	} else {
	    debugString("removeAccessibleSelectionFromContext; ac = null");
	}
    }

    /**
     * select all (if possible) of the children of the AccessibleSelection of an AccessibleContext
     *
     */
    void selectAllAccessibleSelectionFromContext(AccessibleContext ac) {
	if (ac != null) {
            AccessibleSelection as = ac.getAccessibleSelection();
            if (as != null) {
                as.selectAllAccessibleSelection();
            }
	} else {
	    debugString("selectAllAccessibleSelectionFromContext; ac = null");
	}
    }

    // ======== AccessibleTable ========

    Hashtable hashtab = new Hashtable();

    /**
     * returns the AccessibleTable for an AccessibleContext
     */
    public AccessibleTable getAccessibleTableFromContext(AccessibleContext ac) {
	String version = getJavaVersionProperty();
	if ((version != null && version.compareTo("1.3") >= 0)) {
	    if (ac != null) {
		AccessibleTable at = ac.getAccessibleTable();
		if (at != null) {
		    hashtab.put(at, ac);
		    return at;
		}
	    }
	}
	return null;
    }

    /*
     * returns the AccessibleContext that contains an AccessibleTable
     */
    public AccessibleContext getContextFromAccessibleTable(AccessibleTable at) {
	Object o = hashtab.get(at);
	if (o instanceof AccessibleContext) {
	    return (AccessibleContext)o;
	}
	return null;
    }

    /*
     * returns the row count for an AccessibleTable
     */
    public int getAccessibleTableRowCount(AccessibleContext ac) {
	debugString("##### getAccessibleTableRowCount");
	if (ac != null) {
	    AccessibleTable at = ac.getAccessibleTable();
	    if (at != null) {
		int rowCount = at.getAccessibleRowCount();
		debugString("   ##### getAccessibleTableRowCount="+
			    rowCount);
		return rowCount;
	    }
	}
	debugString("   ##### getAccessibleTableRowCount FAILED");
	return -1;
    }

    /*
     * returns the column count for an AccessibleTable
     */
    public int getAccessibleTableColumnCount(AccessibleContext ac) {
	debugString("##### getAccessibleTableColumnCount");
	if (ac != null) {
	    AccessibleTable at = ac.getAccessibleTable();
	    if (at != null) {
		int columnCount = at.getAccessibleColumnCount();
		debugString("   ##### getAccessibleTableColumnCount="
			    +columnCount);
		return columnCount;
	    }
	}
	debugString(" ##### getAccessibleTableColumnCount FAILED");
	return -1;
    }

    /*
     * returns the AccessibleContext for an AccessibleTable cell
     */
    public AccessibleContext getAccessibleTableCellAccessibleContext(AccessibleTable at,
								     int row, int column) {

	debugString("##### getAccessibleTableCellAccessibleContext called");
	debugString("##### -> row = " + row + " column = " + column + " at = " + at);
	if (at != null) {
	    Accessible a = at.getAccessibleAt(row, column);
	    if (a != null) {
		return a.getAccessibleContext();
	    }
	}
	return null;
    }
    
    /*
     * returns the index of a cell at a given row and column in an AccessibleTable
     */
    public int getAccessibleTableCellIndex(AccessibleTable at, int row, int column) {
	debugString("##### getAccessibleTableCellIndex: at="+at);
	if (at != null) {
	    int cellIndex = row * at.getAccessibleColumnCount() + column;
	    debugString("   ##### getAccessibleTableCellIndex="+cellIndex);
	    return cellIndex;
	}
	debugString(" ##### getAccessibleTableCellIndex FAILED");
	return -1;
    }

    /*
     * returns the row extent of a cell at a given row and column in an AccessibleTable
     */
    public int getAccessibleTableCellRowExtent(AccessibleTable at, int row, int column) {
	debugString("##### getAccessibleTableCellRowExtent");
	if (at != null) {
	    int rowExtent = at.getAccessibleRowExtentAt(row, column);
	    debugString("   ##### getAccessibleTableCellRowExtent="+rowExtent);
	    return rowExtent;
	}
	debugString(" ##### getAccessibleTableCellRowExtent FAILED");
	return -1;
    }

    /*
     * returns the column extent of a cell at a given row and column in an AccessibleTable
     */
    public int getAccessibleTableCellColumnExtent(AccessibleTable at, int row, int column) {
	debugString("##### getAccessibleTableCellColumnExtent");
	if (at != null) {
	    int columnExtent = at.getAccessibleColumnExtentAt(row, column);
	    debugString("   ##### getAccessibleTableCellColumnExtent="+columnExtent);
	    return columnExtent;
	}
	debugString(" ##### getAccessibleTableCellColumnExtent FAILED");
	return -1;
    }

    /*
     * returns whether a cell is selected at a given row and column in an AccessibleTable
     */
    public boolean isAccessibleTableCellSelected(AccessibleTable at, int row, 
						 int column) {
	boolean isSelected = false;

	debugString("##### isAccessibleTableCellSelected: ["+row+"]["+column+"]");
	if (at != null) {
       	    Accessible a = at.getAccessibleAt(row, column);
	    if (a != null) {
		AccessibleContext ac = a.getAccessibleContext();
		AccessibleStateSet as = ac.getAccessibleStateSet();
		if (as != null) {
		    isSelected = as.contains(AccessibleState.SELECTED);
		}
	    }
	    debugString(" ##### isAccessibleTableCellSelected="+isSelected);
	    return isSelected;
	}
	debugString(" ##### isAccessibleTableCellSelected FAILED");
	return false;
    }

    /*
     * returns an AccessibleTable that represents the row header in an
     * AccessibleTable
     */
    public AccessibleTable getAccessibleTableRowHeader(AccessibleContext ac) {     

	debugString(" #####  getAccessibleTableRowHeader called");
	if (ac != null) {
	    AccessibleTable at = ac.getAccessibleTable();
	    if (at != null) {
		return at.getAccessibleRowHeader();
	    }
	}
	debugString(" ##### getAccessibleTableRowHeader FAILED");
	return null;
    }

    /*
     * returns an AccessibleTable that represents the column header in an
     * AccessibleTable
     */
    public AccessibleTable getAccessibleTableColumnHeader(AccessibleContext ac) {
	debugString("##### getAccessibleTableColumnHeader");
	if (ac != null) {
	    // workaround for getAccessibleColumnHeader NPE 
	    // when the table header is null
	    Accessible parent = ac.getAccessibleParent();
	    debugString("  parent = "+parent.getClass());
	    if (parent instanceof Accessible) {
		int indexInParent = ac.getAccessibleIndexInParent();
		debugString("  indexInParent = "+indexInParent);
		Accessible child = 
		    parent.getAccessibleContext().getAccessibleChild(indexInParent);
		if (child instanceof JTable) {
		    JTable table = (JTable)child;
		    if (table.getTableHeader() == null) {
			return null;
		    }
		}
	    }
	    AccessibleTable at = ac.getAccessibleTable();
	    if (at != null) {
		return at.getAccessibleColumnHeader();
	    }
	}
	debugString(" ##### getAccessibleTableColumnHeader FAILED");
	return null;
    }

    /*
     * returns the number of row headers in an AccessibleTable that represents
     * the row header in an AccessibleTable
     */
    public int getAccessibleTableRowHeaderRowCount(AccessibleContext ac) {     

	debugString(" #####  getAccessibleTableRowHeaderRowCount called");
	if (ac != null) {
	    AccessibleTable atRowHeader = getAccessibleTableRowHeader(ac);
	    if (atRowHeader != null) {
		return atRowHeader.getAccessibleRowCount();
	    }
	}
	debugString(" ##### getAccessibleTableRowHeaderRowCount FAILED");
	return -1;
    }

    /*
     * returns the number of column headers in an AccessibleTable that represents
     * the row header in an AccessibleTable
     */
    public int getAccessibleTableRowHeaderColumnCount(AccessibleContext ac) {     

	debugString(" #####  getAccessibleTableRowHeaderColumnCount called");
	if (ac != null) {
	    AccessibleTable atRowHeader = getAccessibleTableRowHeader(ac);
	    if (atRowHeader != null) {
		return atRowHeader.getAccessibleColumnCount();
	    }
	}
	debugString(" ##### getAccessibleTableRowHeaderColumnCount FAILED");
	return -1;
    }

    /*
     * returns the number of row headers in an AccessibleTable that represents
     * the column header in an AccessibleTable
     */
    public int getAccessibleTableColumnHeaderRowCount(AccessibleContext ac) {     

	debugString("##### getAccessibleTableColumnHeaderRowCount");
	if (ac != null) {
	    AccessibleTable atColumnHeader = getAccessibleTableColumnHeader(ac);
	    if (atColumnHeader != null) {
		return atColumnHeader.getAccessibleRowCount();
	    }
	}
	debugString(" ##### getAccessibleTableColumnHeaderRowCount FAILED");
	return -1;
    }

    /*
     * returns the number of column headers in an AccessibleTable that represents
     * the column header in an AccessibleTable
     */
    public int getAccessibleTableColumnHeaderColumnCount(AccessibleContext ac) {     

	debugString("#####  getAccessibleTableColumnHeaderColumnCount");
	if (ac != null) {
	    AccessibleTable atColumnHeader = getAccessibleTableColumnHeader(ac);
	    if (atColumnHeader != null) {
		return atColumnHeader.getAccessibleColumnCount();
	    }
	}
	debugString(" ##### getAccessibleTableColumnHeaderColumnCount FAILED");
	return -1;
    }

    /*
     * returns the description of a row header in an AccessibleTable
     */
    public AccessibleContext getAccessibleTableRowDescription(AccessibleTable table,
							      int row) {
	if (table != null) {
	    Accessible a = table.getAccessibleRowDescription(row);
	    if (a != null) {
		return a.getAccessibleContext();
	    }
	}
	return null;
    }

    /*
     * returns the description of a column header in an AccessibleTable
     */
    public AccessibleContext getAccessibleTableColumnDescription(AccessibleTable at,
								 int column) {
	if (at != null) {
	    Accessible a = at.getAccessibleColumnDescription(column);
	    if (a != null) {
		return a.getAccessibleContext();
	    }
	}
	return null;
    }

    /*
     * returns the number of rows selected in an AccessibleTable
     */
    public int getAccessibleTableRowSelectionCount(AccessibleTable at) {
	if (at != null) {
	    int [] selections = at.getSelectedAccessibleRows();
	    return selections.length;
	}
	return -1;
    }

    /*
     * returns the row number of the i-th selected row in an AccessibleTable
     */
    public int getAccessibleTableRowSelections(AccessibleTable at, int i) {
	if (at != null) {
	    int [] selections = at.getSelectedAccessibleRows();
	    if (selections.length > i) {
		return selections[i];
	    }
	}
	return -1;
    }

    /*
     * returns whether a row is selected in an AccessibleTable
     */
    public boolean isAccessibleTableRowSelected(AccessibleTable at, 
						int row) {
	if (at != null) {
	    return at.isAccessibleRowSelected(row);
	}
	return false;
    }

    /*
     * returns whether a column is selected in an AccessibleTable
     */
    public boolean isAccessibleTableColumnSelected(AccessibleTable at, 
						   int column) {
	if (at != null) {
	    return at.isAccessibleColumnSelected(column);
	}
	return false;
    }

    /*
     * returns the number of columns selected in an AccessibleTable
     */
    public int getAccessibleTableColumnSelectionCount(AccessibleTable at) {
	if (at != null) {
	    int [] selections = at.getSelectedAccessibleColumns();
	    return selections.length;
	}
	return -1;
    }

    /*
     * returns the row number of the i-th selected row in an AccessibleTable
     */
    public int getAccessibleTableColumnSelections(AccessibleTable at, int i) {
	if (at != null) {
	    int [] selections = at.getSelectedAccessibleColumns();
	    if (selections.length > i) {
		return selections[i];
	    }
	}
	return -1;
    }

    /* ===== AccessibleExtendedTable (since 1.4) ===== */

    /*
     * returns the row number for a cell at a given index in an AccessibleTable
     */
    public int getAccessibleTableRow(AccessibleTable at, int index) {
	if (! runningOnJDK1_4) {
	    return -1;
	}
	if (at instanceof AccessibleExtendedTable) {
	    return index / at.getAccessibleColumnCount();
	}
	return -1;
    }

    // ===== AccessibleExtendedTable (since 1.4) =====

    /*
     * returns the column number for a cell at a given index in an AccessibleTable
     */
    public int getAccessibleTableColumn(AccessibleTable at, int index) {
	if (! runningOnJDK1_4) {
	    return -1;
	}
	if (at instanceof AccessibleExtendedTable) {
	    return index % at.getAccessibleColumnCount();
	}
	return -1;
    }

    // ===== AccessibleExtendedTable (since 1.4) =====
    /*
     * returns the index for a cell at a given row and column in an 
     * AccessibleTable
     */
    public int getAccessibleTableIndex(AccessibleTable at, int row, int column) {
	if (! runningOnJDK1_4) {
	    return -1;
	}
	if (at instanceof AccessibleExtendedTable) {
	    return row * at.getAccessibleColumnCount() + column;
	}
	return -1;
    }

    // ========== AccessibleRelationSet ==========

    /*
     * returns the number of relations in the AccessibleContext's
     * AccessibleRelationSet
     */
    public int getAccessibleRelationCount(AccessibleContext ac) {
	String version = getJavaVersionProperty();
	if ((version != null && version.compareTo("1.3") >= 0)) {
	    if (ac != null) {
		AccessibleRelationSet ars = ac.getAccessibleRelationSet();
		return ars.size();
	    }
	}
	return 0;
    }

    /*
     * returns the ith relation key in the AccessibleContext's 
     * AccessibleRelationSet
     */
    public String getAccessibleRelationKey(AccessibleContext ac, int i) {
	if (ac != null) {
	    AccessibleRelationSet ars = ac.getAccessibleRelationSet();
	    AccessibleRelation [] relations = ars.toArray();
	    if (i >= 0 && i < relations.length) {
		return relations[i].getKey();
	    }
	}
	return null;
    }

    /*
     * returns the number of targets in a relation in the AccessibleContext's
     * AccessibleRelationSet
     */
    public int getAccessibleRelationTargetCount(AccessibleContext ac, int i) {
	if (ac != null) {
	    AccessibleRelationSet ars = ac.getAccessibleRelationSet();
	    AccessibleRelation [] relations = ars.toArray();
	    if (i >= 0 && i < relations.length) {
		Object [] targets = relations[i].getTarget();
		return targets.length;
	    }
	}
	return -1;
    }

    /*
     * returns the jth target in the ith relation in the AccessibleContext's 
     * AccessibleRelationSet
     */
    public AccessibleContext getAccessibleRelationTarget(AccessibleContext ac, 
							 int i, int j) {
	debugString("***** getAccessibleRelationTarget");
	if (ac != null) {
	    AccessibleRelationSet ars = ac.getAccessibleRelationSet();
	    AccessibleRelation [] relations = ars.toArray();
	    if (i >= 0 && i < relations.length) {
		Object [] targets = relations[i].getTarget();
		if (j >= 0 & j < targets.length) {
		    Object o = targets[j];
		    debugString("***** target="+o.getClass());
		    if (o instanceof Accessible) {
			return ((Accessible)o).getAccessibleContext();
		    }
		}
	    }
	}
	return null;
    }

    // ========= AccessibleHypertext =========

    /*
     * Returns the AccessibleHypertext
     */
    AccessibleHypertext getAccessibleHypertext(AccessibleContext ac) {
	debugString("getAccessibleHyperlink");

	if (ac == null) {
	    return null;
	}
	AccessibleText at = ac.getAccessibleText();
	if (! (at instanceof AccessibleHypertext)) {
	    return null;
	}
	return ((AccessibleHypertext)at);
    }

    /*
     * Returns the number of AccessibleHyperlinks
     */
    int getAccessibleHyperlinkCount(AccessibleContext ac) {
	debugString("getAccessibleHyperlinkCount");
	if (ac == null) {
	    return 0;
	}
	AccessibleHypertext hypertext = getAccessibleHypertext(ac);
	if (hypertext == null) {
	    return 0;
	}
	return hypertext.getLinkCount();
    }

    /*
     * Returns the hyperlink at the specified index
     */
    AccessibleHyperlink getAccessibleHyperlink(AccessibleContext ac, int i) {
	debugString("getAccessibleHyperlink("+i+")");
	if (ac == null) {
	    return null;
	}
	AccessibleHypertext hypertext = getAccessibleHypertext(ac);
	if (hypertext == null) {
	    return null;
	}
	if (i < 0 || i >= hypertext.getLinkCount()) {
	    return null;
	}
	AccessibleHyperlink link = hypertext.getLink(i);
	if (link == null || (! link.isValid())) {
	    return null;
	}
	return link;
    }

    /*
     * Returns the hyperlink object description
     */
    String getAccessibleHyperlinkText(AccessibleContext ac, int i) {
	debugString("getAccessibleHyperlinkText("+i+")");
	AccessibleHyperlink link = getAccessibleHyperlink(ac, i);
	if (link == null) {
	    return null;
	}
	Object o = link.getAccessibleActionObject(0);
	if (o != null) {
	    debugString(" object = "+o);
	    return o.toString();
	} else {
	    return null;
	}
    }

    /*
     * Returns the start index of the hyperlink text
     */
    int getAccessibleHyperlinkStartIndex(AccessibleContext ac, int i) {
	debugString("getAccessibleHyperlinkStartIndex("+i+")");
	AccessibleHyperlink link = getAccessibleHyperlink(ac, i);
	if (link == null) {
	    return -1;
	}
	return link.getStartIndex();
    }

    /*
     * Returns the end index of the hyperlink text
     */
    int getAccessibleHyperlinkEndIndex(AccessibleContext ac, int i) {
	debugString("getAccessibleHyperlinkEndIndex("+i+")");
	AccessibleHyperlink link = getAccessibleHyperlink(ac, i);
	if (link == null) {
	    return -1;
	}
	return link.getEndIndex();
    }

    /*
     * Actives the hyperlink
     */
    boolean activateAccessibleHyperlink(AccessibleContext ac,
					AccessibleHyperlink link) {
	debugString("activateAccessibleHyperlink: link = "+link.getClass());
	if (link == null) {
	    return false;
	}
	boolean retval = link.doAccessibleAction(0);
	debugString("activateAccessibleHyperlink: returning = "+retval);
	return retval;
    }

    // ============ AccessibleKeyBinding =============

    /*
     * return the number of key bindings associated with this context
     */
    int getAccessibleKeyBindingsCount(AccessibleContext ac) {
	debugString("getAccessibleKeyBindingsCount");
	if (ac == null  || (! runningOnJDK1_4) ) {
	    return 0;
	}
	AccessibleComponent comp = ac.getAccessibleComponent();
	if (! (comp instanceof AccessibleExtendedComponent)) {
	    return 0;
	}
	AccessibleExtendedComponent aec = (AccessibleExtendedComponent)comp;
	AccessibleKeyBinding akb = aec.getAccessibleKeyBinding();
	if (akb == null) {
	    return 0;
	}
	debugString("   #key bindings = "+akb.getAccessibleKeyBindingCount());
	return akb.getAccessibleKeyBindingCount();
    }

    /*
     * return key binding character at the specified index
     */
     char getAccessibleKeyBindingChar(AccessibleContext ac, int index) {
	debugString("getAccessibleKeyBindingChar: index = "+index);
	if (ac == null  || (! runningOnJDK1_4) ) {
	    return 0;
	}
	AccessibleComponent comp = ac.getAccessibleComponent();
	if (! (comp instanceof AccessibleExtendedComponent)) {
	    return 0;
	}
	AccessibleExtendedComponent aec = (AccessibleExtendedComponent)comp;
	AccessibleKeyBinding akb = aec.getAccessibleKeyBinding();
	if (akb == null) {
	    return 0;
	}
	Object o = akb.getAccessibleKeyBinding(index);
	if (o instanceof KeyStroke) {
       	    KeyStroke keyStroke = (KeyStroke)o;
	    String keyText = KeyEvent.getKeyText(keyStroke.getKeyCode());
	    debugString("   key text = "+keyText);
	    
	    if (keyText != null || keyText.length() > 0) {
		CharSequence seq = keyText.subSequence(0, 1);
		if (seq != null || seq.length() > 0) {
		    return seq.charAt(0);
		}
	    }
	}
	return 0;
    }

    /*
     * return key binding modifiers at the specified index
     */
     int getAccessibleKeyBindingModifiers(AccessibleContext ac, int index) {
	debugString("getAccessibleKeyBindingModifiers: index = "+index);
	if (ac == null  || (! runningOnJDK1_4) ) {
	    return 0;
	}
	AccessibleComponent comp = ac.getAccessibleComponent();
	if (! (comp instanceof AccessibleExtendedComponent)) {
	    return 0;
	}
	AccessibleExtendedComponent aec = (AccessibleExtendedComponent)comp;
	AccessibleKeyBinding akb = aec.getAccessibleKeyBinding();
	if (akb == null) {
	    return 0;
	}
	Object o = akb.getAccessibleKeyBinding(index);
	if (o instanceof KeyStroke) {
	    debugString("   key binding modifiers = "+((KeyStroke)o).getModifiers());
	    return ((KeyStroke)o).getModifiers();
	}
	return 0;
    }

    // ========== AccessibleIcon ============

    /*
     * return the number of icons associated with this context
     */
    int getAccessibleIconsCount(AccessibleContext ac) {
	debugString("getAccessibleIconsCount");
	if (ac == null) {
	    return 0;
	}
	AccessibleIcon [] ai = ac.getAccessibleIcon();
	if (ai == null) {
	    return 0;
	}
	debugString("   #icons = "+ai.length);
	return ai.length;
    }

    /*
     * return icon description at the specified index
     */
    String getAccessibleIconDescription(AccessibleContext ac, int index) {
	debugString("getAccessibleIconDescription: index = "+index);
	if (ac == null) {
	    return null;
	}
	AccessibleIcon [] ai = ac.getAccessibleIcon();
	if (ai == null || index < 0 || index >= ai.length) {
	    return null;
	}
	debugString("   icon description    = "+
		    ai[index].getAccessibleIconDescription());
	return ai[index].getAccessibleIconDescription();
    }

    /*
     * return icon height at the specified index
     */
    int getAccessibleIconHeight(AccessibleContext ac, int index) {
	debugString("getAccessibleIconHeight: index = "+index);
	if (ac == null) {
	    return 0;
	}
	AccessibleIcon [] ai = ac.getAccessibleIcon();
	if (ai == null || index < 0 || index >= ai.length) {
	    return 0;
	}
	debugString("   icon height    = "+
		    ai[index].getAccessibleIconHeight());
	return ai[index].getAccessibleIconHeight();
    }

    /*
     * return icon width at the specified index
     */
    int getAccessibleIconWidth(AccessibleContext ac, int index) {
	debugString("getAccessibleIconWidth: index = "+index);
	if (ac == null) {
	    return 0;
	}
	AccessibleIcon [] ai = ac.getAccessibleIcon();
	if (ai == null || index < 0 || index >= ai.length) {
	    return 0;
	}
	debugString("   icon height    = "+
		    ai[index].getAccessibleIconWidth());
	return ai[index].getAccessibleIconWidth();
    }

    // ========= AccessibleAction ===========

    /*
     * return the number of icons associated with this context
     */
    int getAccessibleActionsCount(AccessibleContext ac) {
	debugString("getAccessibleActionsCount");
	if (ac == null) {
	    return 0;
	}
	AccessibleAction aa = ac.getAccessibleAction();
	if (aa == null) {
	    return 0;
	}
	debugString("   #actions = "+aa.getAccessibleActionCount());
	return aa.getAccessibleActionCount();
    }

    /*
     * return icon description at the specified index
     */
    String getAccessibleActionName(AccessibleContext ac, int index) {
	debugString("getAccessibleActionName: index = "+index);
	if (ac == null) {
	    return null;
	}
	AccessibleAction aa = ac.getAccessibleAction();
	if (aa == null) {
	    return null;
	}
	debugString("   action name    = "+
		    aa.getAccessibleActionDescription(index));
	return aa.getAccessibleActionDescription(index);
    }
    /*
     * return icon description at the specified index
     */
    boolean doAccessibleActions(AccessibleContext ac, String name) {
	debugString("doAccessibleActions: action name = "+name);
	if (ac == null || name == null) {
	    return false;
	}
	AccessibleAction aa = ac.getAccessibleAction();
	if (aa == null) {
	    return false;
	}
	int index = -1;
	int numActions = aa.getAccessibleActionCount();
	for (int i = 0; i < numActions; i++) {
	    String actionName = aa.getAccessibleActionDescription(i);
	    if (name.equals(actionName)) {
		index = i;
		break;
	    }
	}
	if (index == -1) {
	    debugString("  no such action: "+name);
	    return false;
	}
	debugString(" doing action #"+index+": "+name);
	boolean retval = aa.doAccessibleAction(index);
	debugString(" action "+name+": "+retval);
	return retval;
    }

    /* ===== AT utility methods ===== */

    /**
     * Sets the contents of an AccessibleContext that
     * implements AccessibleEditableText with the
     * specified text string.
     * Returns whether successful.
     */
    boolean setTextContents(AccessibleContext ac, String text) {
	debugString("setTextContents: ac = "+ac+"; text = "+text);

	if (! (ac instanceof AccessibleEditableText)) {
	    debugString("   ac not instanceof AccessibleEditableText: "+ac);
	    return false;
	}
	if (text == null) {
	    debugString("   text is null");
	    return false;
	}

	// check whether the text field is editable
	AccessibleStateSet ass = ac.getAccessibleStateSet();
	if (! ass.contains(AccessibleState.ENABLED)) {
	    debugString("   text is not editable");
	    return false;
	}

	((AccessibleEditableText)ac).setTextContents(text);
	debugString("   setTextContents succeeded");
	return true;
    }

    /**
     * Returns the Accessible Context of an Internal Frame object that is 
     * the ancestor of a given object.  If the object is an Internal Frame 
     * object or an Internal Frame ancestor object was found, returns the
     * object's AccessibleContext.
     * If there is no ancestor object that has an Accessible Role of 
     * Internal Frame, returns (AccessibleContext)0.
     */
    AccessibleContext getInternalFrame (AccessibleContext ac) {
	return getParentWithRole(ac, AccessibleRole.INTERNAL_FRAME.toString());
    }
    
    /**
     * Returns the Accessible Context for the top level object in 
     * a Java Window.  This is same Accessible Context that is obtained 
     * from GetAccessibleContextFromHWND for that window.  Returns 
     * (AccessibleContext)0 on error.
     */
    private AccessibleContext getTopLevelObject (AccessibleContext ac) {
	debugString("getTopLevelObject; ac = "+ac);
	if (ac == null) {
	    return null;
	}
	Accessible parent = ac.getAccessibleParent();
	if (parent == null) {
	    debugString("   getTopLevelObject; returning: "+ac);
	    return ac;
	}
	Accessible tmp = parent;
	while (tmp != null && tmp.getAccessibleContext() != null) {
	    debugString("   getTopLevelObject; next parent = "+
			tmp.getAccessibleContext().getClass());
	    parent = tmp;
	    tmp = parent.getAccessibleContext().getAccessibleParent();
	}
	debugString("   getTopLevelObject; returning: "+
		    parent.getAccessibleContext().getClass());
	return parent.getAccessibleContext();
    }
    
    /**
     * Returns the parent AccessibleContext that has the specified AccessibleRole.  
     * Returns null on error or if the AccessibleContext does not exist.
     */
    public AccessibleContext getParentWithRole (AccessibleContext ac, 
						String roleName) {
	debugString("getParentWithRole; ac = "+ac+"; role = "+roleName);
	if (ac == null || roleName == null) {
	    return null;
	}

	AccessibleRole role = (AccessibleRole)accessibleRoleMap.get(roleName);
	if (role == null) {
	    debugString("   getParentWithRole; no such role: "+roleName);
	    return ac;
	}

	Accessible parent = ac.getAccessibleParent();
	if (parent == null && ac.getAccessibleRole() == role) {
	    debugString("   getParentWithRole; returning ac: "+ac.getClass());
	    return ac;
	}
	
	Accessible tmp = parent;
	AccessibleContext tmp_ac = null;
	
	while (tmp != null && (tmp_ac = tmp.getAccessibleContext()) != null) {
	    debugString("   getParentWithRole; next parent = "+tmp_ac.getClass());
	    AccessibleRole ar = tmp_ac.getAccessibleRole();
	    if (ar == role) {
		// found
		debugString("   getParentWithRole; returning tmp_ac: "+tmp_ac.getClass());
		return tmp_ac;
	    }
	    parent = tmp;
	    tmp = parent.getAccessibleContext().getAccessibleParent();
	}
	// not found
	debugString("   getParentWithRole; parent not found");
	return null;
    }
    
    /**
     * Returns the parent AccessibleContext that has the specified AccessibleRole.  
     * Otherwise, returns the top level object for the Java Window.  
     * Returns (AccessibleContext)0 on error.
     */
    AccessibleContext getParentWithRoleElseRoot (AccessibleContext ac,
						 String roleName) {
	AccessibleContext retval = getParentWithRole(ac, roleName);
	if (retval == null) {
	    retval = getTopLevelObject(ac);
	}
	return retval;
    }
    
    /**
     * Returns how deep in the object hierarchy a given object is.
     * The top most object in the object hierarchy has an object depth of 0.
     * Returns -1 on error.
     */
    int getObjectDepth (AccessibleContext ac) {
	debugString("getObjectDepth: ac = "+ac);
	int count = 0;

	if (ac == null) {
	    return -1;
	}
	Accessible parent = ac.getAccessibleParent();
	if (parent == null) {
	    debugString("   getObjectDepth: count = "+count);
	    return count;
	}
	Accessible tmp = parent;
	while (tmp != null && tmp.getAccessibleContext() != null) {
	    debugString("   getTopLevelObject; next parent = "+
			tmp.getAccessibleContext().getClass());
	    parent = tmp;
	    tmp = parent.getAccessibleContext().getAccessibleParent();
	    count++;
	}
	debugString("   getObjectDepth: count = "+count);
	return count;
    }

    /**
     * Returns the Accessible Context of the current ActiveDescendent of an object.
     * Returns (AccessibleContext)0 on error.
     */
    AccessibleContext getActiveDescendent (AccessibleContext ac) {
	debugString("getActiveDescendent: ac = "+ac);
	if (ac == null) {
	    return null;
	}
	AccessibleSelection as = ac.getAccessibleSelection();
	if (as == null) {
	    debugString("   getActiveDescendent: no selection");
	    return null;
	}
	// assume single selection
	if (as.getAccessibleSelectionCount() != 1) {
	    debugString("   getActiveDescendent: selection count > 1");
	    return null;
	}
	Accessible a = as.getAccessibleSelection(0);
	if (a == null) {
	    debugString("   getActiveDescendent: selection is null");
	    return null;
	}
	debugString("   getActiveDescendent: returning: "+a.getAccessibleContext());
	return a.getAccessibleContext();
    }

    /* ===== Java object memory management code ===== */

    /**
     * Class to track object references to ensure the
     * Java VM doesn't garbage collect them
     */
    protected class ObjectReferences {
	
	
	protected class Reference {
	    public int value;
	    
	    public Reference(int i) {
		value = i;
	    }
	    
	    public String toString() {
		return ("refCount: " + value);
	    }
	}

	/**
	 * table object references, to keep 'em from being garbage collected
	 */
	private Hashtable refs;
		

	/**
	 * Constructor
	 */
	public ObjectReferences() {
	    refs = new Hashtable(4);
	}

	/**
	 * Debugging: dump the contents of ObjectReferences' refs Hashtable
	 */
	private String dump() {
	    return refs.toString();
	}

	/**
	 * Increment ref count; set to 1 if we have no references for it
	 */
	public void increment(Object o) {
	    if (refs.containsKey(o)) {
		((Reference) refs.get(o)).value++;
	    } else {
		refs.put(o, new Reference(1));
	    }
	}

	/**
	 * Decrement ref count; remove if count drops to 0
	 */
	public void decrement(Object o) {
	    Reference aRef = (Reference) refs.get(o);
	    if (aRef != null) {
		aRef.value--;
		if (aRef.value == 0) {
		    refs.remove(o);
		} else if (aRef.value < 0) {
		    debugString("ERROR: decrementing reference count below 0");	
		}
	    } else {
		debugString("ERROR: object to decrement not in ObjectReferences table");
	    }
	}

    }

    /* ===== event handling code ===== */

   /**
     * native method for handling property change events
     */
    public native void propertyCaretChange(PropertyChangeEvent e, 
					   AccessibleContext src,
					   int oldValue, int newValue);
    public native void propertyDescriptionChange(PropertyChangeEvent e, 
						 AccessibleContext src,
						 String oldValue, String newValue);
    public native void propertyNameChange(PropertyChangeEvent e, 
					  AccessibleContext src,
					  String oldValue, String newValue);
    public native void propertySelectionChange(PropertyChangeEvent e, 
					       AccessibleContext src);
    public native void propertyStateChange(PropertyChangeEvent e, 
					   AccessibleContext src,
					   String oldValue, String newValue);
    public native void propertyTextChange(PropertyChangeEvent e, 
					  AccessibleContext src);
    public native void propertyValueChange(PropertyChangeEvent e, 
					   AccessibleContext src,
					   String oldValue, String newValue);
    public native void propertyVisibleDataChange(PropertyChangeEvent e, 
						 AccessibleContext src);
    public native void propertyChildChange(PropertyChangeEvent e, 
					   AccessibleContext src,
					   AccessibleContext oldValue,
					   AccessibleContext newValue);
    public native void propertyActiveDescendentChange(PropertyChangeEvent e, 
						      AccessibleContext src,
						      AccessibleContext oldValue,
						      AccessibleContext newValue);


    /**
     * native methods for handling focus events
     */
    public native void focusGained(FocusEvent e, AccessibleContext src);
    public native void focusLost(FocusEvent e, AccessibleContext src);

    /**
     * native method for handling caret events
     */
    public native void caretUpdate(CaretEvent e, AccessibleContext src);

    /**
     * native methods for handling mouse events
     */
    public native void mouseClicked(MouseEvent e, AccessibleContext src);
    public native void mouseEntered(MouseEvent e, AccessibleContext src);
    public native void mouseExited(MouseEvent e, AccessibleContext src);
    public native void mousePressed(MouseEvent e, AccessibleContext src);
    public native void mouseReleased(MouseEvent e, AccessibleContext src);

    /**
     * native methods for handling menu & popupMenu events
     */
    public native void menuCanceled(MenuEvent e, AccessibleContext src);
    public native void menuDeselected(MenuEvent e, AccessibleContext src);
    public native void menuSelected(MenuEvent e, AccessibleContext src);
    public native void popupMenuCanceled(PopupMenuEvent e, AccessibleContext src);
    public native void popupMenuWillBecomeInvisible(PopupMenuEvent e, 
						    AccessibleContext src);
    public native void popupMenuWillBecomeVisible(PopupMenuEvent e, 
						  AccessibleContext src);

    /* ===== event definitions ===== */

    public static final long PROPERTY_CHANGE_EVENTS = 1;
    public static final long FOCUS_GAINED_EVENTS = 2;
    public static final long FOCUS_LOST_EVENTS = 4;
    public static final long FOCUS_EVENTS = (FOCUS_GAINED_EVENTS | FOCUS_LOST_EVENTS);

    public static final long CARET_UPATE_EVENTS = 8;
    public static final long CARET_EVENTS = CARET_UPATE_EVENTS;

    public static final long MOUSE_CLICKED_EVENTS = 16;
    public static final long MOUSE_ENTERED_EVENTS = 32;
    public static final long MOUSE_EXITED_EVENTS = 64;
    public static final long MOUSE_PRESSED_EVENTS = 128;
    public static final long MOUSE_RELEASED_EVENTS = 256;
    public static final long MOUSE_EVENTS = (MOUSE_CLICKED_EVENTS | MOUSE_ENTERED_EVENTS |
                                             MOUSE_EXITED_EVENTS | MOUSE_PRESSED_EVENTS |
                                             MOUSE_RELEASED_EVENTS);

    public static final long MENU_CANCELED_EVENTS = 512;
    public static final long MENU_DESELECTED_EVENTS = 1024;
    public static final long MENU_SELECTED_EVENTS = 2048;
    public static final long MENU_EVENTS = (MENU_CANCELED_EVENTS | MENU_DESELECTED_EVENTS |
                                            MENU_SELECTED_EVENTS);

    public static final long POPUPMENU_CANCELED_EVENTS = 4096;
    public static final long POPUPMENU_WILL_BECOME_INVISIBLE_EVENTS = 8192;
    public static final long POPUPMENU_WILL_BECOME_VISIBLE_EVENTS = 16384;
    public static final long POPUPMENU_EVENTS = (POPUPMENU_CANCELED_EVENTS |
						 POPUPMENU_WILL_BECOME_INVISIBLE_EVENTS |
						 POPUPMENU_WILL_BECOME_VISIBLE_EVENTS);

    /* These use their own numbering scheme, to ensure sufficient expansion room */
    public static final long PROPERTY_NAME_CHANGE_EVENTS = 1;
    public static final long PROPERTY_DESCRIPTION_CHANGE_EVENTS = 2;
    public static final long PROPERTY_STATE_CHANGE_EVENTS = 4;
    public static final long PROPERTY_VALUE_CHANGE_EVENTS = 8;
    public static final long PROPERTY_SELECTION_CHANGE_EVENTS = 16;
    public static final long PROPERTY_TEXT_CHANGE_EVENTS = 32;
    public static final long PROPERTY_CARET_CHANGE_EVENTS = 64;
    public static final long PROPERTY_VISIBLEDATA_CHANGE_EVENTS = 128;
    public static final long PROPERTY_CHILD_CHANGE_EVENTS = 256;
    public static final long PROPERTY_ACTIVEDESCENDENT_CHANGE_EVENTS = 512;


    public static final long PROPERTY_EVENTS = (PROPERTY_NAME_CHANGE_EVENTS |
                                                PROPERTY_DESCRIPTION_CHANGE_EVENTS |
                                                PROPERTY_STATE_CHANGE_EVENTS |
                                                PROPERTY_VALUE_CHANGE_EVENTS |
                                                PROPERTY_SELECTION_CHANGE_EVENTS |
                                                PROPERTY_TEXT_CHANGE_EVENTS |
                                                PROPERTY_CARET_CHANGE_EVENTS |
                                                PROPERTY_VISIBLEDATA_CHANGE_EVENTS |
                                                PROPERTY_CHILD_CHANGE_EVENTS |
                                                PROPERTY_ACTIVEDESCENDENT_CHANGE_EVENTS);

    /**
     * The EventHandler class listens for Java events and
     * forwards them to the AT
     */
    protected class EventHandler implements PropertyChangeListener,
					    FocusListener, CaretListener, 
					    MenuListener, PopupMenuListener, 
					    MouseListener, WindowListener, 
					    ChangeListener {

        AccessBridge accessBridge;
        long javaEventMask = 0;
        long accessibilityEventMask = 0;

        public EventHandler(AccessBridge bridge) {
            accessBridge = bridge;
	    
	    // Register to receive WINDOW_OPENED and WINDOW_CLOSED
	    // events.  Add the event source as a native window
	    // handler is it implements NativeWindowHandler.
	    // SwingEventMonitor.addWindowListener(this);
        }

        // --------- Event Notification Registration methods

	/**
	 * Invoked the first time a window is made visible.
	 */
	public void windowOpened(WindowEvent e) {
	    // If the window is a NativeWindowHandler, add it.
	    Object o = e.getSource();
	    if (o instanceof NativeWindowHandler) {
		addNativeWindowHandler((NativeWindowHandler)o);
	    }
	}
	
	/**
	 * Invoked when the user attempts to close the window
	 * from the window's system menu.  If the program does not 
	 * explicitly hide or dispose the window while processing 
	 * this event, the window close operation will be cancelled.
	 */
	public void windowClosing(WindowEvent e) {}
	
	/**
	 * Invoked when a window has been closed as the result
	 * of calling dispose on the window.
	 */
	public void windowClosed(WindowEvent e) {
	    // If the window is a NativeWindowHandler, remove it.
	    Object o = e.getSource();
	    if (o instanceof NativeWindowHandler) {
		removeNativeWindowHandler((NativeWindowHandler)o);
	    }
	}
	
	/**
	 * Invoked when a window is changed from a normal to a
	 * minimized state. For many platforms, a minimized window 
	 * is displayed as the icon specified in the window's 
	 * iconImage property.
	 * @see java.awt.Frame#setIconImage
	 */
	public void windowIconified(WindowEvent e) {}
	
	/**
	 * Invoked when a window is changed from a minimized
	 * to a normal state.
	 */
	public void windowDeiconified(WindowEvent e) {}
	
	/**
	 * Invoked when the Window is set to be the active Window. Only a Frame or
	 * a Dialog can be the active Window. The native windowing system may
	 * denote the active Window or its children with special decorations, such
	 * as a highlighted title bar. The active Window is always either the
	 * focused Window, or the first Frame or Dialog that is an owner of the
	 * focused Window.
	 */
	public void windowActivated(WindowEvent e) {}
	
	/**
	 * Invoked when a Window is no longer the active Window. Only a Frame or a
	 * Dialog can be the active Window. The native windowing system may denote
	 * the active Window or its children with special decorations, such as a
	 * highlighted title bar. The active Window is always either the focused
	 * Window, or the first Frame or Dialog that is an owner of the focused
	 * Window.
	 */
	public void windowDeactivated(WindowEvent e) {}
	
        /**
         * Turn on event monitoring for the event type passed in
         * If necessary, add the appropriate event listener (if
         * no other event of that type is being listened for)
         */
        public void addJavaEventNotification(long type) {
            long newEventMask = javaEventMask | type;
	    /*
	      if (((javaEventMask & PROPERTY_EVENTS) == 0) &&
	      ((newEventMask & PROPERTY_EVENTS) != 0)) {
	      AccessibilityEventMonitor.addPropertyChangeListener(this);
	      }
	    */
            if (((javaEventMask & FOCUS_EVENTS) == 0) &&
                ((newEventMask & FOCUS_EVENTS) != 0)) {
                SwingEventMonitor.addFocusListener(this);
            }
            if (((javaEventMask & CARET_EVENTS) == 0) &&
                ((newEventMask & CARET_EVENTS) != 0)) {
                SwingEventMonitor.addCaretListener(this);
            }
            if (((javaEventMask & MOUSE_EVENTS) == 0) &&
                ((newEventMask & MOUSE_EVENTS) != 0)) {
                SwingEventMonitor.addMouseListener(this);
            }
            if (((javaEventMask & MENU_EVENTS) == 0) &&
                ((newEventMask & MENU_EVENTS) != 0)) {
                SwingEventMonitor.addMenuListener(this);
                SwingEventMonitor.addPopupMenuListener(this);
            }
            if (((javaEventMask & POPUPMENU_EVENTS) == 0) &&
                ((newEventMask & POPUPMENU_EVENTS) != 0)) {
                SwingEventMonitor.addPopupMenuListener(this);
            }

            javaEventMask = newEventMask;
        }

        /**
         * Turn off event monitoring for the event type passed in
         * If necessary, remove the appropriate event listener (if
         * no other event of that type is being listened for)
         */
        public void removeJavaEventNotification(long type) {
            long newEventMask = javaEventMask & (~type);
	    /*
	      if (((javaEventMask & PROPERTY_EVENTS) != 0) &&
	      ((newEventMask & PROPERTY_EVENTS) == 0)) {
	      AccessibilityEventMonitor.removePropertyChangeListener(this);
	      }
	    */
            if (((javaEventMask & FOCUS_EVENTS) != 0) &&
                ((newEventMask & FOCUS_EVENTS) == 0)) {
                SwingEventMonitor.removeFocusListener(this);
            }
            if (((javaEventMask & CARET_EVENTS) != 0) &&
                ((newEventMask & CARET_EVENTS) == 0)) {
                SwingEventMonitor.removeCaretListener(this);
            }
            if (((javaEventMask & MOUSE_EVENTS) == 0) &&
                ((newEventMask & MOUSE_EVENTS) != 0)) {
                SwingEventMonitor.removeMouseListener(this);
            }
            if (((javaEventMask & MENU_EVENTS) == 0) &&
                ((newEventMask & MENU_EVENTS) != 0)) {
                SwingEventMonitor.removeMenuListener(this);
            }
            if (((javaEventMask & POPUPMENU_EVENTS) == 0) &&
                ((newEventMask & POPUPMENU_EVENTS) != 0)) {
                SwingEventMonitor.removePopupMenuListener(this);
            }

            javaEventMask = newEventMask;
        }

        /**
         * Turn on event monitoring for the event type passed in
         * If necessary, add the appropriate event listener (if
         * no other event of that type is being listened for)
         */
        public void addAccessibilityEventNotification(long type) {
            long newEventMask = accessibilityEventMask | type;
            if (((accessibilityEventMask & PROPERTY_EVENTS) == 0) &&
                ((newEventMask & PROPERTY_EVENTS) != 0)) {
                AccessibilityEventMonitor.addPropertyChangeListener(this);
            }
            accessibilityEventMask = newEventMask;
        }

        /**
         * Turn off event monitoring for the event type passed in
         * If necessary, remove the appropriate event listener (if
         * no other event of that type is being listened for)
         */
        public void removeAccessibilityEventNotification(long type) {
            long newEventMask = accessibilityEventMask & (~type);
            if (((accessibilityEventMask & PROPERTY_EVENTS) != 0) &&
                ((newEventMask & PROPERTY_EVENTS) == 0)) {
                AccessibilityEventMonitor.removePropertyChangeListener(this);
            }
            accessibilityEventMask = newEventMask;
        }

	/**
	 *  ------- property change event glue
	 */
        public void propertyChange(PropertyChangeEvent e) {

            accessBridge.debugString("propertyChange(" + e.toString() + ") called");

	    if ((accessibilityEventMask & PROPERTY_EVENTS) != 0) {
                Object o = e.getSource();
                AccessibleContext ac;

                if (o instanceof AccessibleContext) {
                    ac = (AccessibleContext) o;
                } else {
                    Accessible a = Translator.getAccessible(e.getSource());
                    ac = a.getAccessibleContext();
                }
                if (ac != null) {

                    String propertyName = e.getPropertyName();

                    if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_CARET_PROPERTY) == 0) {
                        int oldValue = 0;
                        int newValue = 0;

                        if (e.getOldValue() instanceof Integer) {
                            oldValue = ((Integer) e.getOldValue()).intValue();
                        } 
                        if (e.getNewValue() instanceof Integer) {
                            newValue = ((Integer) e.getNewValue()).intValue();
                        } 
                        accessBridge.debugString(" - about to call propertyCaretChange()");
                        accessBridge.propertyCaretChange(e, ac, oldValue, newValue);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_DESCRIPTION_PROPERTY) == 0) {
                        String oldValue = null;
                        String newValue = null;

                        if (e.getOldValue() != null) {
                            oldValue = e.getOldValue().toString();
                        } 
                        if (e.getNewValue() != null) {
                            newValue = e.getNewValue().toString();
                        } 
                        accessBridge.debugString(" - about to call propertyDescriptionChange()");
                        accessBridge.propertyDescriptionChange(e, ac, oldValue, newValue);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_NAME_PROPERTY) == 0) {
                        String oldValue = null;
                        String newValue = null;

                        if (e.getOldValue() != null) {
                            oldValue = e.getOldValue().toString();
                        } 
                        if (e.getNewValue() != null) {
                            newValue = e.getNewValue().toString();
                        } 
                        accessBridge.debugString(" - about to call propertyNameChange()");
                        accessBridge.propertyNameChange(e, ac, oldValue, newValue);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_SELECTION_PROPERTY) == 0) {
                        accessBridge.debugString(" - about to call propertySelectionChange()");
                        accessBridge.propertySelectionChange(e, ac);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_STATE_PROPERTY) == 0) {
                        String oldValue = null;
                        String newValue = null;
			/*
			  if (e.getOldValue() != null) {
			  oldValue = e.getOldValue().toString();
			  } 
			  if (e.getNewValue() != null) {
			  newValue = e.getNewValue().toString();
			  }
			*/

			// Localization fix requested by Oliver for EA-1
			if (e.getOldValue() != null) {
			    AccessibleState oldState = (AccessibleState) e.getOldValue();
			    oldValue = oldState.toDisplayString(Locale.US);
			}
			if (e.getNewValue() != null) {
			    AccessibleState newState = (AccessibleState) e.getNewValue();
			    newValue = newState.toDisplayString(Locale.US);
			} 

                        accessBridge.debugString(" - about to call propertyStateChange()");
                        accessBridge.propertyStateChange(e, ac, oldValue, newValue);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_TEXT_PROPERTY) == 0) {
                        accessBridge.debugString(" - about to call propertyTextChange()");
                        accessBridge.propertyTextChange(e, ac);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_VALUE_PROPERTY) == 0) {  // strings 'cause of floating point, etc.
                        String oldValue = null;
                        String newValue = null;

                        if (e.getOldValue() != null) {
                            oldValue = e.getOldValue().toString();
                        } 
                        if (e.getNewValue() != null) {
                            newValue = e.getNewValue().toString();
                        } 
                        accessBridge.debugString(" - about to call propertyDescriptionChange()");
                        accessBridge.propertyValueChange(e, ac, oldValue, newValue);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_VISIBLE_DATA_PROPERTY) == 0) {
                        accessBridge.propertyVisibleDataChange(e, ac);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_CHILD_PROPERTY) == 0) {
                        AccessibleContext oldAC = null;
                        AccessibleContext newAC = null;
                        Accessible a;

                        if (e.getOldValue() instanceof AccessibleContext) {
                            oldAC = (AccessibleContext) e.getOldValue();
                        }                        
                        if (e.getNewValue() instanceof AccessibleContext) {
                            newAC = (AccessibleContext) e.getNewValue();
                        }
                        accessBridge.debugString(" - about to call propertyChildChange()");
                        accessBridge.propertyChildChange(e, ac, oldAC, newAC);

                    } else if (propertyName.compareTo(AccessibleContext.ACCESSIBLE_ACTIVE_DESCENDANT_PROPERTY) == 0) {
                        AccessibleContext oldAC = null;
                        AccessibleContext newAC = null;
                        Accessible a;

                        if (e.getOldValue() instanceof Accessible) {
                            oldAC = ((Accessible) e.getOldValue()).getAccessibleContext();
                        } else if (e.getOldValue() instanceof Component) {
                            a = Translator.getAccessible((Component) e.getOldValue());
                            if (a != null) {
                                oldAC = a.getAccessibleContext();
                            }
                        }
                        
                        if (e.getNewValue() instanceof Accessible) {
                            newAC = ((Accessible) e.getNewValue()).getAccessibleContext();
                        } else if (e.getNewValue() instanceof Component) {
                            a = Translator.getAccessible((Component) e.getNewValue());
                            if (a != null) {
                                oldAC = a.getAccessibleContext();
                            }
                        }
                        accessBridge.debugString("  - about to call propertyActiveDescendentChange()");
                        accessBridge.propertyActiveDescendentChange(e, ac, oldAC, newAC);
                    }
                }
            }
        }

	/**
	 *  ------- focus event glue
	 */
	private boolean stateChangeListenerAdded = false;

        public void focusGained(FocusEvent e) {
	    if (runningOnJDK1_4) {
		processFocusGained();
		
	    } else {
		if ((javaEventMask & FOCUS_GAINED_EVENTS) != 0) {
		    Accessible a = Translator.getAccessible(e.getSource());
		    if (a != null) {
			accessBridge.focusGained(e, a.getAccessibleContext());
		    }
		}
	    }
	}

	public void stateChanged(ChangeEvent e) {
	    processFocusGained();
	}

        private void processFocusGained() {
	    Component focusOwner = KeyboardFocusManager.
		getCurrentKeyboardFocusManager().getFocusOwner();
	    if (focusOwner == null) {
		return;
	    }
	    
	    // Only menus and popup selections are handled by the JRootPane.
	    if (focusOwner instanceof JRootPane) {
		MenuElement [] path = 
		    MenuSelectionManager.defaultManager().getSelectedPath();
		if (path.length > 1) {
		    Component penult = path[path.length-2].getComponent();
		    Component last = path[path.length-1].getComponent();
		    
		    if (last instanceof JPopupMenu) {
			// This is a popup with nothing in the popup
			// selected. The menu itself is selected.
			FocusEvent e = 
			    new FocusEvent(penult, FocusEvent.FOCUS_GAINED);
			accessBridge.focusGained(e, ((Accessible)penult).getAccessibleContext());
		    } else if (penult instanceof JPopupMenu) {
			// This is a popup with an item selected
			FocusEvent e = 
			    new FocusEvent(last, FocusEvent.FOCUS_GAINED);
			accessBridge.focusGained(e, ((Accessible)last).getAccessibleContext());
		    }
		}
	    } else {
		// The focus owner has the selection.
		if (focusOwner instanceof Accessible) {
		    FocusEvent e = new FocusEvent(focusOwner, 
						  FocusEvent.FOCUS_GAINED);
		    accessBridge.focusGained(e, ((Accessible)focusOwner).getAccessibleContext());
		}
	    }
        }

        public void focusLost(FocusEvent e) {
	    if ((javaEventMask & FOCUS_LOST_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.focusLost(e, a.getAccessibleContext());
                }
            }
        }

	/**
	 *  ------- caret event glue
	 */
        public void caretUpdate(CaretEvent e) {
	    if ((javaEventMask & CARET_UPATE_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.caretUpdate(e, a.getAccessibleContext());
                }
            }
        }

	/**
	 *  ------- mouse event glue
	 */

        public void mouseClicked(MouseEvent e) {
	    if ((javaEventMask & MOUSE_CLICKED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.mouseClicked(e, a.getAccessibleContext());
                }
            }
        }

        public void mouseEntered(MouseEvent e) {
	    if ((javaEventMask & MOUSE_ENTERED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.mouseEntered(e, a.getAccessibleContext());
                }
            }
        }

        public void mouseExited(MouseEvent e) {
	    if ((javaEventMask & MOUSE_EXITED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.mouseExited(e, a.getAccessibleContext());
                }
            }
        }

        public void mousePressed(MouseEvent e) {
	    if ((javaEventMask & MOUSE_PRESSED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.mousePressed(e, a.getAccessibleContext());
                }
            }
        }

        public void mouseReleased(MouseEvent e) {
	    if ((javaEventMask & MOUSE_RELEASED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.mouseReleased(e, a.getAccessibleContext());
                }
            }
        }

	/**
	 *  ------- menu event glue
	 */
        public void menuCanceled(MenuEvent e) {
	    if ((javaEventMask & MENU_CANCELED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.menuCanceled(e, a.getAccessibleContext());
                }
            }
        }

        public void menuDeselected(MenuEvent e) {
	    if ((javaEventMask & MENU_DESELECTED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.menuDeselected(e, a.getAccessibleContext());
                }
            }
        }

        public void menuSelected(MenuEvent e) {
	    if ((javaEventMask & MENU_SELECTED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.menuSelected(e, a.getAccessibleContext());
                }
            }
        }

        public void popupMenuCanceled(PopupMenuEvent e) {
	    if ((javaEventMask & POPUPMENU_CANCELED_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.popupMenuCanceled(e, a.getAccessibleContext());
                }
            }
        }

        public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
	    if ((javaEventMask & POPUPMENU_WILL_BECOME_INVISIBLE_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.popupMenuWillBecomeInvisible(e, a.getAccessibleContext());
                }
            }
        }

        public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
	    if ((javaEventMask & POPUPMENU_WILL_BECOME_VISIBLE_EVENTS) != 0) {
                Accessible a = Translator.getAccessible(e.getSource());
                if (a != null) {
                    accessBridge.popupMenuWillBecomeVisible(e, a.getAccessibleContext());
                }
            }
        }
    }

    // --------- Event Notification Registration methods

    /**
     *  Wrapper method around eventHandler.addJavaEventNotification()
     */
    public void addJavaEventNotification(long type) {
        eventHandler.addJavaEventNotification(type);
    }

    /**
     *  Wrapper method around eventHandler.removeJavaEventNotification()
     */
    public void removeJavaEventNotification(long type) {
        eventHandler.removeJavaEventNotification(type);
    }


    /**
     *  Wrapper method around eventHandler.addAccessibilityEventNotification()
     */
    public void addAccessibilityEventNotification(long type) {
        eventHandler.addAccessibilityEventNotification(type);
    }

    /**
     *  Wrapper method around eventHandler.removeAccessibilityEventNotification()
     */
    public void removeAccessibilityEventNotification(long type) {
        eventHandler.removeAccessibilityEventNotification(type);
    }

    /**
     ******************************************************
     * All AccessibleRoles
     *
     * We shouldn't have to do this since it requires us
     * to synchronize the allAccessibleRoles array when
     * the AccessibleRoles class interface changes. However,
     * there is no Accessibility API method to get all 
     * AccessibleRoles
     ******************************************************
     */
    private AccessibleRole [] allAccessibleRoles = {
	/**
	 * Object is used to alert the user about something.
	 */
	AccessibleRole.ALERT,

	/**
	 * The header for a column of data.
	 */
	AccessibleRole.COLUMN_HEADER,

	/** 
	 * Object that can be drawn into and is used to trap
	 * events.
	 * @see #FRAME
	 * @see #GLASS_PANE
	 * @see #LAYERED_PANE
	 */     
	AccessibleRole.CANVAS,

	/** 
	 * A list of choices the user can select from.  Also optionally 
	 * allows the user to enter a choice of their own.
	 */     
	AccessibleRole.COMBO_BOX,

	/** 
	 * An iconified internal frame in a DESKTOP_PANE.
	 * @see #DESKTOP_PANE
	 * @see #INTERNAL_FRAME
	 */
	AccessibleRole.DESKTOP_ICON,

	/** 
	 * A frame-like object that is clipped by a desktop pane.  The
	 * desktop pane, internal frame, and desktop icon objects are 
	 * often used to create multiple document interfaces within an
	 * application.
	 * @see #DESKTOP_ICON
	 * @see #DESKTOP_PANE
	 * @see #FRAME
	 */
	AccessibleRole.INTERNAL_FRAME,

	/**
	 * A pane that supports internal frames and 
	 * iconified versions of those internal frames.
	 * @see #DESKTOP_ICON
	 * @see #INTERNAL_FRAME
	 */
	AccessibleRole.DESKTOP_PANE,

	/** 
	 * A specialized pane whose primary use is inside a DIALOG
	 * @see #DIALOG
	 */
	AccessibleRole.OPTION_PANE,

	/** 
	 * A top level window with no title or border.
	 * @see #FRAME
	 * @see #DIALOG
	 */
	AccessibleRole.WINDOW,

	/**
	 * A top level window with a title bar, border, menu bar, etc.  It is
	 * often used as the primary window for an application.
	 * @see #DIALOG
	 * @see #CANVAS
	 * @see #WINDOW
	 */
	AccessibleRole.FRAME,

	/** 
	 * A top level window with title bar and a border.  A dialog is similar 
	 * to a frame, but it has fewer properties and is often used as a 
	 * secondary window for an application.
	 * @see #FRAME
	 * @see #WINDOW
	 */
	AccessibleRole.DIALOG,

	/**
	 * A specialized dialog that lets the user choose a color.
	 */
	AccessibleRole.COLOR_CHOOSER,


	/**
	 * A pane that allows the user to navigate through 
	 * and select the contents of a directory.  May be used
	 * by a file chooser.
	 * @see #FILE_CHOOSER
	 */
	AccessibleRole.DIRECTORY_PANE,

	/**
	 * A specialized dialog that displays the files in the directory
	 * and lets the user select a file, browse a different directory,
	 * or specify a filename.  May use the directory pane to show the
	 * contents of a directory.
	 * @see #DIRECTORY_PANE
	 */
	AccessibleRole.FILE_CHOOSER,

	/** 
	 * An object that fills up space in a user interface.  It is often
	 * used in interfaces to tweak the spacing between components,
	 * but serves no other purpose.
	 */
	AccessibleRole.FILLER,

	/**
	 * A hypertext anchor
	 */
	// AccessibleRole.HYPERLINK,

	/**
	 * A small fixed size picture, typically used to decorate components.
	 */
	AccessibleRole.ICON,

	/** 
	 * An object used to present an icon or short string in an interface.
	 */
	AccessibleRole.LABEL,

	/**
	 * A specialized pane that has a glass pane and a layered pane as its
	 * children.
	 * @see #GLASS_PANE
	 * @see #LAYERED_PANE
	 */
	AccessibleRole.ROOT_PANE,

	/**
	 * A pane that is guaranteed to be painted on top
	 * of all panes beneath it.
	 * @see #ROOT_PANE
	 * @see #CANVAS
	 */
	AccessibleRole.GLASS_PANE,

	/** 
	 * A specialized pane that allows its children to be drawn in layers,
	 * providing a form of stacking order.  This is usually the pane that
	 * holds the menu bar as well as the pane that contains most of the
	 * visual components in a window.
	 * @see #GLASS_PANE
	 * @see #ROOT_PANE
	 */
	AccessibleRole.LAYERED_PANE,

	/**
	 * An object that presents a list of objects to the user and allows the
	 * user to select one or more of them.  A list is usually contained
	 * within a scroll pane.
	 * @see #SCROLL_PANE
	 * @see #LIST_ITEM
	 */
	AccessibleRole.LIST,

	/**
	 * An object that presents an element in a list.  A list is usually
	 * contained within a scroll pane.
	 * @see #SCROLL_PANE
	 * @see #LIST
	 */
	AccessibleRole.LIST_ITEM,

	/**
	 * An object usually drawn at the top of the primary dialog box of
	 * an application that contains a list of menus the user can choose
	 * from.  For example, a menu bar might contain menus for "File,"
	 * "Edit," and "Help."
	 * @see #MENU
	 * @see #POPUP_MENU
	 * @see #LAYERED_PANE
	 */
	AccessibleRole.MENU_BAR,

	/** 
	 * A temporary window that is usually used to offer the user a 
	 * list of choices, and then hides when the user selects one of
	 * those choices.
	 * @see #MENU
	 * @see #MENU_ITEM
	 */
	AccessibleRole.POPUP_MENU,

	/** 
	 * An object usually found inside a menu bar that contains a list
	 * of actions the user can choose from.  A menu can have any object
	 * as its children, but most often they are menu items, other menus,
	 * or rudimentary objects such as radio buttons, check boxes, or
	 * separators.  For example, an application may have an "Edit" menu 
	 * that contains menu items for "Cut" and "Paste."
	 * @see #MENU_BAR
	 * @see #MENU_ITEM
	 * @see #SEPARATOR
	 * @see #RADIO_BUTTON
	 * @see #CHECK_BOX
	 * @see #POPUP_MENU
	 */    
	AccessibleRole.MENU,

	/**
	 * An object usually contained in a menu that presents an action 
	 * the user can choose.  For example, the "Cut" menu item in an
	 * "Edit" menu would be an action the user can select to cut the
	 * selected area of text in a document.
	 * @see #MENU_BAR
	 * @see #SEPARATOR
	 * @see #POPUP_MENU
	 */
	AccessibleRole.MENU_ITEM,

	/**
	 * An object usually contained in a menu to provide a visual
	 * and logical separation of the contents in a menu.  For example,
	 * the "File" menu of an application might contain menu items for
	 * "Open," "Close," and "Exit," and will place a separator between
	 * "Close" and "Exit" menu items.
	 * @see #MENU
	 * @see #MENU_ITEM
	 */
	AccessibleRole.SEPARATOR,

	/**
	 * An object that presents a series of panels (or page tabs), one at a 
	 * time, through some mechanism provided by the object.  The most common 
	 * mechanism is a list of tabs at the top of the panel.  The children of
	 * a page tab list are all page tabs.
	 * @see #PAGE_TAB
	 */
	AccessibleRole.PAGE_TAB_LIST,

	/**
	 * An object that is a child of a page tab list.  Its sole child is
	 * the panel that is to be presented to the user when the user 
	 * selects the page tab from the list of tabs in the page tab list.
	 * @see #PAGE_TAB_LIST
	 */
	AccessibleRole.PAGE_TAB,

	/**
	 * A generic container that is often used to group objects.
	 */
	AccessibleRole.PANEL,

	/**
	 * An object used to indicate how much of a task has been completed.
	 */
	AccessibleRole.PROGRESS_BAR,

	/**
	 * A text object used for passwords, or other places where the 
	 * text contents is not shown visibly to the user
	 */
	AccessibleRole.PASSWORD_TEXT,

	/**
	 * An object the user can manipulate to tell the application to do
	 * something.
	 * @see #CHECK_BOX
	 * @see #TOGGLE_BUTTON
	 * @see #RADIO_BUTTON
	 */
	AccessibleRole.PUSH_BUTTON,
  
	/**
	 * A specialized push button that can be checked or unchecked, but
	 * does not provide a separate indicator for the current state.
	 * @see #PUSH_BUTTON
	 * @see #CHECK_BOX
	 * @see #RADIO_BUTTON
	 */
	AccessibleRole.TOGGLE_BUTTON,

	/**
	 * A choice that can be checked or unchecked and provides a 
	 * separate indicator for the current state.
	 * @see #PUSH_BUTTON
	 * @see #TOGGLE_BUTTON
	 * @see #RADIO_BUTTON
	 */
	AccessibleRole.CHECK_BOX,

	/**
	 * A specialized check box that will cause other radio buttons in the
	 * same group to become unchecked when this one is checked.  
	 * @see #PUSH_BUTTON
	 * @see #TOGGLE_BUTTON
	 * @see #CHECK_BOX
	 */
	AccessibleRole.RADIO_BUTTON,

	/**
	 * The header for a row of data.
	 */
	AccessibleRole.ROW_HEADER,

	/**
	 * An object that allows a user to incrementally view a large amount
	 * of information.  Its children can include scroll bars and a viewport.
	 * @see #SCROLL_BAR
	 * @see #VIEWPORT
	 */
	AccessibleRole.SCROLL_PANE,

	/** 
	 * An object usually used to allow a user to incrementally view a
	 * large amount of data.  Usually used only by a scroll pane.
	 * @see #SCROLL_PANE
	 */
	AccessibleRole.SCROLL_BAR,

	/**
	 * An object usually used in a scroll pane.  It represents the portion 
	 * of the entire data that the user can see.  As the user manipulates 
	 * the scroll bars, the contents of the viewport can change.
	 * @see #SCROLL_PANE
	 */
	AccessibleRole.VIEWPORT,

	/**
	 * An object that allows the user to select from a bounded range.  For
	 * example, a slider might be used to select a number between 0 and 100.
	 */    
	AccessibleRole.SLIDER,

	/**
	 * A specialized panel that presents two other panels at the same time.
	 * Between the two panels is a divider the user can manipulate to make
	 * one panel larger and the other panel smaller.
	 */
	AccessibleRole.SPLIT_PANE,

	/**
	 * An object used to present information in terms of rows and columns.
	 * An example might include a spreadsheet application.
	 */
	AccessibleRole.TABLE,

	/**
	 * An object that presents text to the user.  The text is usually
	 * editable by the user as opposed to a label.
	 * @see #LABEL
	 */
	AccessibleRole.TEXT,

	/**
	 * An object used to present hierarchical information to the user.
	 * The individual nodes in the tree can be collapsed and expanded
	 * to provide selective disclosure of the tree's contents.
	 */
	AccessibleRole.TREE,

	/**
	 * A bar or palette usually composed of push buttons or toggle buttons.
	 * It is often used to provide the most frequently used functions for an
	 * application.
	 */
	AccessibleRole.TOOL_BAR,

	/** 
	 * An object that provides information about another object.  The 
	 * accessibleDescription property of the tool tip is often displayed 
	 * to the user in a small "help bubble" when the user causes the 
	 * mouse to hover over the object associated with the tool tip.
	 */
	AccessibleRole.TOOL_TIP,

	/**
	 * An AWT component, but nothing else is known about it.
	 * @see #SWING_COMPONENT
	 * @see #UNKNOWN
	 */
	AccessibleRole.AWT_COMPONENT,

	/**
	 * A Swing component, but nothing else is known about it.
	 * @see #AWT_COMPONENT
	 * @see #UNKNOWN
	 */
	AccessibleRole.SWING_COMPONENT,

	/**
	 * The object contains some Accessible information, but its role is
	 * not known.
	 * @see #AWT_COMPONENT
	 * @see #SWING_COMPONENT
	 */
	AccessibleRole.UNKNOWN,

	// These roles are only available in JDK 1.4

	/**
	 * A STATUS_BAR is an simple component that can contain
	 * multiple labels of status information to the user.
	 AccessibleRole.STATUS_BAR,
    
	 /**
	 * A DATE_EDITOR is a component that allows users to edit
	 * java.util.Date and java.util.Time objects
	 AccessibleRole.DATE_EDITOR,
    
	 /**
	 * A SPIN_BOX is a simple spinner component and its main use
	 * is for simple numbers.
	 AccessibleRole.SPIN_BOX,
    
	 /**
	 * A FONT_CHOOSER is a component that lets the user pick various
	 * attributes for fonts.
	 AccessibleRole.FONT_CHOOSER,
    
	 /**
	 * A GROUP_BOX is a simple container that contains a border
	 * around it and contains components inside it.
	 AccessibleRole.GROUP_BOX

	 /**
	 * Since JDK 1.5
	 *
	 * A text header

	 AccessibleRole.HEADER,
 
	 /**
	 * A text footer

	 AccessibleRole.FOOTER,
 
	 /**
	 * A text paragraph

	 AccessibleRole.PARAGRAPH,
 
	 /**
	 * A ruler is an object used to measure distance

	 AccessibleRole.RULER,
 
	 /**
	 * A role indicating the object acts as a formula for
	 * calculating a value.  An example is a formula in
	 * a spreadsheet cell.
	 AccessibleRole.EDITBAR
	*/
    };
}
 
